
# Imports
import argparse
import logging
import threading
import time


import pysailfish.internal.observability.log_helper as lh
from pysailfish.internal.Network.TCPCast import TCPCast, TCPStatus

# Variables
server_ip: str = "127.0.0.1" # "0.0.0.0" --- for linux and docker "127.0.0.1" --- for windows
server_port: int = 23456
tcp_client: TCPCast = TCPCast()

# Functions
def init_argparse(description: str = "") -> argparse.Namespace:
    """! initial argument parser

    @param description Argument parser description

    @return argument parser
    """
    parser = argparse.ArgumentParser(description=description)
    # parser.add_argument("-a", "--action", required=True)
    # parser.add_argument("-l", "--lang", required=True)
    # parser.add_argument("-pn", "--projectName", default="")
    return parser.parse_args()

def client_worker() -> int:
    logger = lh.init_logger(logger_name="sailfish_logger", is_json_output=False)
    if tcp_client.InitComponent(if_address=server_ip
                                , if_port=server_port
                                , is_client=True) == TCPStatus.ERROR:
        logger.error("Cannot create tcp client. Stop.")
        return -1

    if tcp_client.Connect(to_address=server_ip
                          , to_port=server_port) == TCPStatus.ERROR:
        logger.error("Cannot connect to tcp server. Stop.")
        return -1

    count = 0
    while count < 10000:
        count += 1
        recv_status, recv_msg = tcp_client.ClientRecv()
        if recv_status == TCPStatus.SUCCESS:
            if recv_msg != "":
                logger.info(f"Client received message: {recv_msg}")

    return 0

def main() -> None:
    """! sailfish main function

    @return None
    """
    logger = lh.init_logger(logger_name="sailfish_logger", is_json_output=False)
    args = init_argparse(description="sailfish Inputs")
    logger.info(f"We get args: {args}")

    # Here put your logic
    client_th: threading.Thread = threading.Thread(target=client_worker)
    client_th.start()

    time.sleep(5)

if __name__ == "__main__":
    main()



from typing import Dict, List, Set, Tuple, Union, Any

import pysailfish.internal.MT_EA.mt4_const as mc
from pysailfish.internal.DataCtrl.MT4DataCtrl.MT4DataCtrl import MT4DataCtrl

# https://docs.mql4.com/indicators
class technical_indicators(object):
    def __init__(self):
        self.__data_ctrl: MT4DataCtrl = None
        self.__fun_num_map: Dict[str, int] = dict()

    def init_component(self, data_ctrl: MT4DataCtrl, fun_num_map: Dict[str, int]):
        self.__data_ctrl = data_ctrl
        self.__fun_num_map = fun_num_map

    def i_ac(self, symbol: str, timeframe: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iAC"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_ad(self, symbol: str, timeframe: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iAD"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_adx(self, symbol: str, timeframe: int, period: int, applied_price: int, mode: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iADX"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(period))
        self.__data_ctrl.add_data_to_msg(data=str(applied_price))
        self.__data_ctrl.add_data_to_msg(data=str(mode))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_alligator(self
                    , symbol: str
                    , timeframe: int
                    , jaw_period: int
                    , jaw_shift: int
                    , teeth_period: int
                    , teeth_shift: int
                    , lips_period: int
                    , lips_shift: int
                    , ma_method: int
                    , applied_price: int
                    , mode: int
                    , shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iAlligator"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(jaw_period))
        self.__data_ctrl.add_data_to_msg(data=str(jaw_shift))
        self.__data_ctrl.add_data_to_msg(data=str(teeth_period))
        self.__data_ctrl.add_data_to_msg(data=str(teeth_shift))
        self.__data_ctrl.add_data_to_msg(data=str(lips_period))
        self.__data_ctrl.add_data_to_msg(data=str(lips_shift))
        self.__data_ctrl.add_data_to_msg(data=str(ma_method))
        self.__data_ctrl.add_data_to_msg(data=str(applied_price))
        self.__data_ctrl.add_data_to_msg(data=str(mode))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_ao(self, symbol: str, timeframe: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iAO"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_atr(self, symbol: str, timeframe: int, period: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iATR"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(period))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_bears_power(self, symbol: str, timeframe: int, period: int, applied_price: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iBearsPower"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(period))
        self.__data_ctrl.add_data_to_msg(data=str(applied_price))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_bands(self, symbol: str, timeframe: int, period: int, deviation: float, bands_shift: int, applied_price: int, mode: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iBands"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(period))
        self.__data_ctrl.add_data_to_msg(data=str(deviation))
        self.__data_ctrl.add_data_to_msg(data=str(bands_shift))
        self.__data_ctrl.add_data_to_msg(data=str(applied_price))
        self.__data_ctrl.add_data_to_msg(data=str(mode))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_bands_on_array(self, array: List[float], total: int, period: int, deviation: float, bands_shift: int, applied_price: int, mode: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iBandsOnArray"]))
        self.__data_ctrl.add_data_to_msg(data=str(len(array)))
        self.__data_ctrl.add_data_to_msg(data=",".join([str(ele) for ele in array]))
        self.__data_ctrl.add_data_to_msg(data=str(total))
        self.__data_ctrl.add_data_to_msg(data=str(period))
        self.__data_ctrl.add_data_to_msg(data=str(deviation))
        self.__data_ctrl.add_data_to_msg(data=str(bands_shift))
        self.__data_ctrl.add_data_to_msg(data=str(applied_price))
        self.__data_ctrl.add_data_to_msg(data=str(mode))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_bulls_power(self, symbol: str, timeframe: int, period: int, applied_price: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iBullsPower"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(period))
        self.__data_ctrl.add_data_to_msg(data=str(applied_price))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_cci(self, symbol: str, timeframe: int, period: int, applied_price: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iCCI"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(period))
        self.__data_ctrl.add_data_to_msg(data=str(applied_price))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_cci_on_array(self, array: List[float], total: int, period: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iCCIOnArray"]))
        self.__data_ctrl.add_data_to_msg(data=str(len(array)))
        self.__data_ctrl.add_data_to_msg(data=",".join([str(ele) for ele in array]))
        self.__data_ctrl.add_data_to_msg(data=str(total))
        self.__data_ctrl.add_data_to_msg(data=str(period))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_custom(self, **kwargs) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iCustom"]))
        for k, v in kwargs.items():
            self.__data_ctrl.add_data_to_msg(data=v)
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_de_marker(self, symbol: str, timeframe: int, period: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iDeMarker"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(period))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_envelopes(self
                    , symbol: str
                    , timeframe: int
                    , ma_period: int
                    , ma_method: int
                    , ma_shift: int
                    , applied_price: int
                    , deviation: float
                    , mode: int
                    , shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iEnvelopes"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(ma_period))
        self.__data_ctrl.add_data_to_msg(data=str(ma_method))
        self.__data_ctrl.add_data_to_msg(data=str(ma_shift))
        self.__data_ctrl.add_data_to_msg(data=str(applied_price))
        self.__data_ctrl.add_data_to_msg(data=str(deviation))
        self.__data_ctrl.add_data_to_msg(data=str(mode))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_envelopes_on_array(self
                             , array: List[float]
                             , total: int
                             , ma_period: int
                             , ma_method: int
                             , ma_shift: int
                             , deviation: float
                             , mode: int
                             , shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iEnvelopesOnArray"]))
        self.__data_ctrl.add_data_to_msg(data=str(len(array)))
        self.__data_ctrl.add_data_to_msg(data=",".join([str(ele) for ele in array]))
        self.__data_ctrl.add_data_to_msg(data=str(total))
        self.__data_ctrl.add_data_to_msg(data=str(ma_period))
        self.__data_ctrl.add_data_to_msg(data=str(ma_method))
        self.__data_ctrl.add_data_to_msg(data=str(ma_shift))
        self.__data_ctrl.add_data_to_msg(data=str(deviation))
        self.__data_ctrl.add_data_to_msg(data=str(mode))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_force(self, symbol: str, timeframe: int, period: int, ma_method: int, applied_price: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iForce"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(period))
        self.__data_ctrl.add_data_to_msg(data=str(ma_method))
        self.__data_ctrl.add_data_to_msg(data=str(applied_price))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_fratals(self, symbol: str, timeframe: int, mode: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iFratals"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(mode))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_gator(self
                , symbol: str
                , timeframe: int
                , jaw_period: int
                , jaw_shift: int
                , teeth_period: int
                , teeth_shift: int
                , lips_period: int
                , lips_shift: int
                , ma_method: int
                , applied_price: int
                , mode: int
                , shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iGator"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(jaw_period))
        self.__data_ctrl.add_data_to_msg(data=str(jaw_shift))
        self.__data_ctrl.add_data_to_msg(data=str(teeth_period))
        self.__data_ctrl.add_data_to_msg(data=str(teeth_shift))
        self.__data_ctrl.add_data_to_msg(data=str(lips_period))
        self.__data_ctrl.add_data_to_msg(data=str(lips_shift))
        self.__data_ctrl.add_data_to_msg(data=str(ma_method))
        self.__data_ctrl.add_data_to_msg(data=str(applied_pr))
        self.__data_ctrl.add_data_to_msg(data=str(mode))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_ichimoku(self, symbol: str, timeframe: int, tenkan_sen: int, kijun_sen: int, senkou_span_b: int, mode: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iIchimoku"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(tenkan_sen))
        self.__data_ctrl.add_data_to_msg(data=str(kijun_sen))
        self.__data_ctrl.add_data_to_msg(data=str(senkou_span_b))
        self.__data_ctrl.add_data_to_msg(data=str(mode))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_bwmfi(self, symbol: str, timeframe: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iBWMFI"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_momentum(self, symbol: str, timeframe: int, period: int, applied_price: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iMomentum"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(period))
        self.__data_ctrl.add_data_to_msg(data=str(applied_price))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_momentum_on_array(self, array: List[float], total: int, period: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iMomentumOnArray"]))
        self.__data_ctrl.add_data_to_msg(data=str(len(array)))
        self.__data_ctrl.add_data_to_msg(data=",".join([str(ele) for ele in array]))
        self.__data_ctrl.add_data_to_msg(data=str(total))
        self.__data_ctrl.add_data_to_msg(data=str(period))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_mfi(self, symbol: str, timeframe: int, period: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iMFI"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(period))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_ma(self, symbol: str, timeframe: int, ma_period: int, ma_shift: int, ma_method: int, applied_price: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iMA"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(ma_period))
        self.__data_ctrl.add_data_to_msg(data=str(ma_shift))
        self.__data_ctrl.add_data_to_msg(data=str(ma_method))
        self.__data_ctrl.add_data_to_msg(data=str(applied_price))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_ma_on_array(self, array: List[float], total: int, ma_period: int, ma_shift: int, ma_method: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iMAOnArray"]))
        self.__data_ctrl.add_data_to_msg(data=str(len(array)))
        self.__data_ctrl.add_data_to_msg(data=",".join([str(ele) for ele in array]))
        self.__data_ctrl.add_data_to_msg(data=str(total))
        self.__data_ctrl.add_data_to_msg(data=str(ma_period))
        self.__data_ctrl.add_data_to_msg(data=str(ma_shift))
        self.__data_ctrl.add_data_to_msg(data=str(ma_method))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_os_ma(self, symbol: str, timeframe: int, fast_ema_period: int, slow_ema_period: int, signal_period: int, applied_price: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iOsMA"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(fast_ema_period))
        self.__data_ctrl.add_data_to_msg(data=str(slow_ema_shift))
        self.__data_ctrl.add_data_to_msg(data=str(signal_period))
        self.__data_ctrl.add_data_to_msg(data=str(applied_price))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_macd(self, symbol: str, timeframe: int, fast_ema_period: int, slow_ema_period: int, signal_period: int, applied_price: int, mode: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iMACD"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(fast_ema_period))
        self.__data_ctrl.add_data_to_msg(data=str(slow_ema_shift))
        self.__data_ctrl.add_data_to_msg(data=str(signal_period))
        self.__data_ctrl.add_data_to_msg(data=str(applied_price))
        self.__data_ctrl.add_data_to_msg(data=str(mode))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_obv(self, symbol: str, timeframe: int, applied_price: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iOBV"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(applied_price))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_sar(self, symbol: str, timeframe: int, step: float, maximum: float, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iSAR"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(step))
        self.__data_ctrl.add_data_to_msg(data=str(maximum))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_rsi(self, symbol: str, timeframe: int, period: int, applied_price: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iRSI"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(period))
        self.__data_ctrl.add_data_to_msg(data=str(applied_price))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_rsi_on_array(self, array: List[float], total: int, period: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iRSIOnArray"]))
        self.__data_ctrl.add_data_to_msg(data=str(len(array)))
        self.__data_ctrl.add_data_to_msg(data=",".join([str(ele) for ele in array]))
        self.__data_ctrl.add_data_to_msg(data=str(total))
        self.__data_ctrl.add_data_to_msg(data=str(period))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_rvi(self, symbol: str, timeframe: int, period: int, mode: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iRVI"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(period))
        self.__data_ctrl.add_data_to_msg(data=str(mode))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_std_dev(self, symbol: str, timeframe: int, ma_period: int, ma_shift: int, ma_method: int, applied_price: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iStdDev"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(ma_period))
        self.__data_ctrl.add_data_to_msg(data=str(ma_shift))
        self.__data_ctrl.add_data_to_msg(data=str(ma_method))
        self.__data_ctrl.add_data_to_msg(data=str(applied_price))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_std_dev_on_array(self, array: List[float], total: int, ma_period: int, ma_shift: int, ma_method: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iStdDevOnArray"]))
        self.__data_ctrl.add_data_to_msg(data=str(len(array)))
        self.__data_ctrl.add_data_to_msg(data=",".join([str(ele) for ele in array]))
        self.__data_ctrl.add_data_to_msg(data=str(total))
        self.__data_ctrl.add_data_to_msg(data=str(ma_period))
        self.__data_ctrl.add_data_to_msg(data=str(ma_shift))
        self.__data_ctrl.add_data_to_msg(data=str(ma_method))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_stochastic(self, symbol: str, timeframe: int, Kperiod: int, Dperiod: int, slowing, int, method: int, price_field: int, mode: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iStochastic"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(Kperiod))
        self.__data_ctrl.add_data_to_msg(data=str(Dperiod))
        self.__data_ctrl.add_data_to_msg(data=str(slowing))
        self.__data_ctrl.add_data_to_msg(data=str(method))
        self.__data_ctrl.add_data_to_msg(data=str(price_field))
        self.__data_ctrl.add_data_to_msg(data=str(mode))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def i_wpr(self, symbol: str, timeframe: int, period: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iWPR"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(period))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()



from typing import List
from datetime import datetime

digits: int = 0
point: float = 0
last_error: int = 0
period: int = 0
symbol: str = ""
ask: float = 0
bars: int = 0
bid: float = 0

magic_num: int = 0

is_connected: bool = False
is_optimization: bool = False
is_testing: bool = False
is_trade_allowed: bool = False

close: List[float] = list()
high: List[float] = list()
low: List[float] = list()
open: List[float] = list()
time: List[datetime] = list()
volume: List[int] = list()



from typing import Dict, List, Set, Tuple, Union, Any
from datetime import datetime
import logging
import time
import decimal

from pysailfish.internal.DataCtrl.MT4DataCtrl.MT4DataCtrl import MT4DataCtrl
from pysailfish.internal.Network.TCPCast import TCPCast, TCPStatus
from pysailfish.internal.MT_EA.mt4_fun_num_map import mt4_fun_num_map
import pysailfish.internal.observability.log_helper as lh
import pysailfish.internal.MT_EA.predefined_variables as pv
import pysailfish.internal.MT_EA.trade_functions as tf
import pysailfish.internal.MT_EA.common_functions as cf
import pysailfish.internal.MT_EA.technical_indicators as ti
import pysailfish.internal.MT_EA.account_information as ai
import pysailfish.internal.MT_EA.market_info as mi
import pysailfish.internal.MT_EA.timeseries_and_indicators_access as taia
import pysailfish.internal.MT_EA.object_functions as of
import pysailfish.internal.MT_EA.chart_operations as co

class MT4_EA(object):
    def __init__(self):
        self._server_ip: str = ""
        self._server_port: int = 0
        self._ea_name: str = ""
        self._is_deinit: bool = False
        self._user_inputs: Dict[str, str] = dict() # key: variable name, value: variable
        self._logger: logging.Logger = None
        self._mt4_server: TCPCast = None

        self.__fun_map: Dict[str, Any] = self.__create_fun_map()
        self.__fun_num_map: Dict[str, int] = mt4_fun_num_map
        self.__data_ctrl: MT4DataCtrl = None

        self._pv = pv # predefined variables namespace
        self._tf = tf.trade_functions()
        self._cf = cf.common_functions()
        self._ti = ti.technical_indicators()
        self._ai = ai.account_information()
        self._mi = mi.market_info()
        self._taia = taia.timeseries_and_indicators_access()
        self._of = of.object_functions()
        self._co = co.chart_operations()

    def InitComponent(self
                      , server_ip: str
                      , server_port: int
                      , ea_name: str):
        self._server_ip = server_ip
        self._server_port = server_port
        self._ea_name = ea_name

        self._logger = lh.init_logger(logger_name=f"{ea_name}_logger", is_json_output=False)

        # tcp cast client
        tcp_cast: TCPCast = TCPCast()
        if tcp_cast.InitComponent(if_address=self._server_ip
                                    , if_port=self._server_port
                                    , is_client=True) == TCPStatus.ERROR:
            self._logger.error("Cannot create tcp client. Stop.")
            exit(1)

        if tcp_cast.Connect(to_address=self._server_ip
                              , to_port=self._server_port) == TCPStatus.ERROR:
            self._logger.error("Cannot connect to tcp server. Stop.")
            exit(1)

        # data control
        self.__data_ctrl: MT4DataCtrl = MT4DataCtrl()
        self.__data_ctrl.init_component(logger=self._logger
                                        , tcp_client=tcp_cast
                                        , mt4_msg_callback=self.mt4_msg_callback)

        # mt3 python interface funtions init
        self._tf.init_component(data_ctrl=self.__data_ctrl
                                , fun_num_map=self.__fun_num_map)
        self._cf.init_component(data_ctrl=self.__data_ctrl
                                , fun_num_map=self.__fun_num_map)
        self._ti.init_component(data_ctrl=self.__data_ctrl
                                , fun_num_map=self.__fun_num_map)
        self._ai.init_component(data_ctrl=self.__data_ctrl
                                , fun_num_map=self.__fun_num_map)
        self._mi.init_component(data_ctrl=self.__data_ctrl
                                , fun_num_map=self.__fun_num_map)
        self._taia.init_component(data_ctrl=self.__data_ctrl
                                  , fun_num_map=self.__fun_num_map)
        self._of.init_component(data_ctrl=self.__data_ctrl
                                , fun_num_map=self.__fun_num_map)
        self._co.init_component(data_ctrl=self.__data_ctrl
                                , fun_num_map=self.__fun_num_map)

    def StartEA(self) -> None:
        self.__data_ctrl.start_th()

        while True:
            time.sleep(1)
            if not self.__data_ctrl.is_connection_okay() or self.IsDeinited():
                self._logger.error(f"Connection lost. ip: {self._server_ip} port: {self._server_port}")
                self.__data_ctrl.stop_th()
                exit(1)

    def OnInit(self, parts: List[str]) -> int:
        self._pv.magic_num = int(parts[0])
        # init user inputs map
        self._user_inputs = dict()
        shift_pos = 1
        for i in range(int(parts[0 + shift_pos])):
            variable_name: str = parts[i*2+1 + shift_pos]
            variable_value: str = parts[i*2+2 + shift_pos]
            self._user_inputs[variable_name] = variable_value

        return self._OnInit()

    def OnDeinit(self, parts: List[str]) -> int:
        self._is_deinit = True
        self._OnDeinit(reason=int(parts[0]))

        # this value have no meaning for mt4 framework, this is a notification of the finish of this function
        return 0

    def OnTick(self, parts: List[str]) -> int:
        self.__UpdatePredefiniedVariables(parts=parts)
        self._OnTick()

        # this value have no meaning for mt4 framework, this is a notification of the finish of this function
        return 0

    def OnTimer(self, parts: List[str]) -> None:
        self._OnTimer()

    def OnTester(self, parts: List[str]) -> float:
        return self._OnTester()

    def OnChartEvent(self, parts: List[str]) -> int:
        self._OnChartEvent(id=int(parts[0])
                           , lparam=int(parts[1])
                           , dparam=float(parts[2])
                           , sparam=parts[3])

        # this value have no meaning for mt4 framework, this is a notification of the finish of this function
        return 0

    def IsDeinited(self) -> bool:
        return self._is_deinit

    def mt4_msg_callback(self, fun_name: str, parts: List[str]):
        # call back to python ea
        fun_ret: Any = self.__fun_map[fun_name](parts=parts)

        if fun_ret is not None:
            # call back to mt4 ea
            self.__data_ctrl.add_data_to_msg(str(self.__fun_num_map[fun_name]))
            self.__data_ctrl.add_data_to_msg(str(fun_ret))
            self.__data_ctrl.send_message()

    # override
    def _OnInit(self) -> int:
        raise NotImplementedError

    # override
    def _OnDeinit(self, reason: int) -> None:
        raise NotImplementedError

    # override
    def _OnTick(self) -> None:
        raise NotImplementedError

    # override
    def _OnTimer(self) -> None:
        raise NotImplementedError

    # override
    def _OnTester(self) -> float:
        raise NotImplementedError

    # override
    def _OnChartEvent(self
                     , id: int
                     , lparam: int
                     , dparam: float
                     , sparam: str) -> None:
        raise NotImplementedError

    # mt4 chart operations
    def ChartApplyTemplate(self, **kwargs) -> bool:
        return None

    def ChartSaveTemplate(self, **kwargs) -> bool:
        return None

    def ChartWindowFind_001(self, **kwargs) -> int:
        return None

    def ChartWindowFind_002(self, **kwargs) -> int:
        return None

    def ChartTimePriceToXY(self, **kwargs) -> bool:
        return None

    def ChartXYToTimePrice(self, **kwargs) -> bool:
        return None

    def ChartOpen(self, **kwargs) -> int:
        return None

    def ChartFirst(self) -> int:
        return None

    def ChartNext(self, **kwargs) -> int:
        return None

    def ChartClose(self, **kwargs) -> bool:
        return None

    def ChartSymbol(self, **kwargs) -> str:
        return None

    def ChartPeriod(self, **kwargs) -> int:
        return None

    def ChartRedraw(self, **kwargs) -> None:
        return None

    def ChartSetDouble(self, **kwargs) -> bool:
        return None

    def ChartSetInteger_001(self, **kwargs) -> bool:
        return None

    def ChartSetInteger_002(self, **kwargs) -> bool:
        return None

    def ChartSetString(self, **kwargs) -> bool:
        return None

    def ChartGetDouble_001(self, **kwargs) -> float:
        return None

    def ChartGetDouble_002(self, **kwargs) -> bool:
        return None

    def ChartGetInteger_001(self, **kwargs) -> int:
        return None

    def ChartGetInteger_002(self, **kwargs) -> bool:
        return None

    def ChartGetString_001(self, **kwargs) -> str:
        return None

    def ChartGetString_002(self, **kwargs) -> bool:
        return None

    def ChartNavigate(self, **kwargs) -> bool:
        return None

    def ChartID(self) -> int:
        return None

    def ChartIndicatorDelete(self, **kwargs) -> bool:
        return None

    def ChartIndicatorName(self, **kwargs) -> str:
        return None

    def ChartIndicatorsTotal(self, **kwargs) -> int:
        return None

    def ChartWindowOnDropped(self) -> int:
        return None

    def ChartPriceOnDropped(self) -> float:
        return None

    def ChartTimeOnDropped(self) -> datetime:
        return None

    def ChartXOnDropped(self) -> int:
        return None

    def ChartYOnDropped(self) -> int:
        return None

    def ChartSetSymbolPeriod(self, **kwargs) -> bool:
        return self._co.chart_set_symbol_period(**kwargs)

    def ChartScreenShot(self, **kwargs) -> bool:
        return None

    def Period(self) -> int:
        return None

    def Symbol(self) -> str:
        return None

    def WindowBarsPerChart(self) -> int:
        return None

    def WindowExpertName(self) -> str:
        return None

    def WindowFind(self, **kwargs) -> int:
        return None

    # mt4 interface functions
    def OrderClose(self, **kwargs) -> bool:
        return self._tf.order_close(**kwargs)

    def OrderCloseBy(self, **kwargs) -> bool:
        return self._tf.order_close_by(**kwargs)

    def OrderClosePrice(self) -> float:
        return self._cf.get_float_from_mt(fun_name="OrderClosePrice")

    def OrderCloseTime(self) -> datetime:
        return self._cf.get_datetime_from_mt(fun_name="OrderCloseTime")

    def OrderComment(self) -> str:
        return self._cf.get_str_from_mt(fun_name="OrderComment")

    def OrderCommission(self) -> float:
        return self._cf.get_float_from_mt(fun_name="OrderCommission")

    def OrderDelete(self, **kwargs) -> bool:
        return self._tf.order_delete(**kwargs)

    def OrderExpiration(self) -> datetime:
        return self._cf.get_datetime_from_mt(fun_name="OrderExpiration")

    def OrderLots(self) -> float:
        return self._cf.get_float_from_mt(fun_name="OrderLots")

    def OrderMagicNumber(self) -> int:
        return self._cf.get_int_from_mt(fun_name="OrderMagicNumber")

    def OrderModify(self, **kwargs) -> bool:
        return self._tf.order_modify(**kwargs)

    def OrderOpenPrice(self) -> float:
        return self._cf.get_float_from_mt(fun_name="OrderOpenPrice")

    def OrderOpenTime(self) -> datetime:
        return self._cf.get_datetime_from_mt(fun_name="OrderOpenTime")

    def OrderPrint(self) -> None:
        return self._tf.order_print()

    def OrderProfit(self) -> float:
        return self._cf.get_float_from_mt(fun_name="OrderProfit")

    def OrderSelect(self, **kwargs) -> bool:
        return self._tf.order_select(**kwargs)

    def OrderSend(self, **kwargs) -> int:
        return self._tf.order_send(**kwargs)

    def OrdersHistoryTotal(self) -> int:
        return self._cf.get_int_from_mt(fun_name="OrdersHistoryTotal")

    def OrderStopLoss(self) -> float:
        return self._cf.get_float_from_mt(fun_name="OrderStopLoss")

    def OrdersTotal(self) -> int:
        return self._cf.get_int_from_mt(fun_name="OrdersTotal")

    def OrderSwap(self) -> float:
        return self._cf.get_float_from_mt(fun_name="OrderSwap")

    def OrderSymbol(self) -> str:
        return self._cf.get_str_from_mt(fun_name="OrderSymbol")

    def OrderTakeProfit(self) -> float:
        return self._cf.get_float_from_mt(fun_name="OrderTakeProfit")

    def OrderTicket(self) -> int:
        return self._cf.get_int_from_mt(fun_name="OrderTicket")

    def OrderType(self) -> int:
        return self._cf.get_int_from_mt(fun_name="OrderType")

    # object functions
    # limit at 10 extra input pairs
    def ObjectCreate_001(self, **kwargs) -> bool:
        return self._of.object_create_001(**kwargs)

    def ObjectCreate_002(self, **kwargs) -> bool:
        return self._of.object_create_002(**kwargs)

    def ObjectName(self, **kwargs) -> str:
        return self._of.object_name(**kwargs)

    def ObjectDelete_001(self, **kwargs) -> bool:
        return self._of.object_delete_001(**kwargs)

    def ObjectDelete_002(self, **kwargs) -> bool:
        return self._of.object_delete_002(**kwargs)

    def ObjectsDeleteAll_001(self, **kwargs) -> int:
        return self._of.objects_delete_all_001(**kwargs)

    def ObjectsDeleteAll_002(self, **kwargs) -> int:
        return self._of.objects_delete_all_002(**kwargs)

    def ObjectsDeleteAll_003(self, **kwargs) -> int:
        return self._of.objects_delete_all_003(**kwargs)

    def ObjectFind_001(self, **kwargs) -> int:
        return self._of.object_find_001(**kwargs)

    def ObjectFind_002(self, **kwargs) -> int:
        return self._of.object_find_002(**kwargs)

    def ObjectGetTimeByValue(self, **kwargs) -> datetime:
        return self._of.object_get_time_by_value(**kwargs)

    def ObjectGetValueByTime(self, **kwargs) -> float:
        return self._of.object_get_value_by_time(**kwargs)

    def ObjectMove(self, **kwargs) -> bool:
        return self._of.object_move(**kwargs)

    def ObjectsTotal_001(self, **kwargs) -> int:
        return self._of.objects_total_001(**kwargs)

    def ObjectsTotal_002(self, **kwargs) -> int:
        return self._of.objects_total_002(**kwargs)

    def ObjectGetDouble(self, **kwargs) -> float:
        return self._of.object_get_double(**kwargs)

    def ObjectGetInteger(self, **kwargs) -> int:
        return self._of.object_get_integer(**kwargs)

    def ObjectGetString(self, **kwargs) -> str:
        return self._of.object_get_string(**kwargs)

    def ObjectSetDouble_001(self, **kwargs) -> bool:
        return self._of.object_set_double_001(**kwargs)

    def ObjectSetDouble_002(self, **kwargs) -> bool:
        return self._of.object_set_double_002(**kwargs)

    def ObjectSetInteger_001(self, **kwargs) -> bool:
        return self._of.object_set_integer_001(**kwargs)

    def ObjectSetInteger_002(self, **kwargs) -> bool:
        return self._of.object_set_integer_002(**kwargs)

    def ObjectSetString_001(self, **kwargs) -> bool:
        return self._of.object_set_string_001(**kwargs)

    def ObjectSetString_002(self, **kwargs) -> bool:
        return self._of.object_set_string_002(**kwargs)

    def TextSetFont(self, **kwargs) -> bool:
        return self._of.text_set_font(**kwargs)

    # def TextOut(self, **kwargs) -> bool:
    #     return self._of.text_out(**kwargs)

    def TextGetSize(self, **kwargs) -> Union[int, int]:
        return self._of.text_get_size(**kwargs)

    def ObjectDescription(self, **kwargs) -> str:
        return self._of.object_description(**kwargs)

    def ObjectGet(self, **kwargs) -> float:
        return self._of.object_get(**kwargs)

    def ObjectGetFiboDescription(self, **kwargs) -> str:
        return self._of.object_get_fibo_description(**kwargs)

    def ObjectGetShiftByValue(self, **kwargs) -> int:
        return self._of.object_get_shift_by_value(**kwargs)

    def ObjectGetValueByShift(self, **kwargs) -> float:
        return self._of.object_get_value_by_shift(**kwargs)

    def ObjectSet(self, **kwargs) -> bool:
        return self._of.object_set(**kwargs)

    def ObjectSetFiboDescription(self, **kwargs) -> bool:
        return self._of.object_set_fibo_description(**kwargs)

    def ObjectSetText(self, **kwargs) -> bool:
        return self._of.object_set_text(**kwargs)

    def ObjectType(self, **kwargs) -> int:
        return self._of.object_type(**kwargs)

    # mt4 technical indicator functions
    def iAC(self, **kwargs) -> float:
        return self._ti.i_ac(**kwargs)

    def iAD(self, **kwargs) -> float:
        return self._ti.i_ad(**kwargs)

    def iADX(self, **kwargs) -> float:
        return self._ti.i_adx(**kwargs)

    def iAlligator(self, **kwargs) -> float:
        return self._ti.i_alligator(**kwargs)

    def iAO(self, **kwargs) -> float:
        return self._ti.i_ao(**kwargs)

    def iATR(self, **kwargs) -> float:
        return self._ti.i_atr(**kwargs)

    def iBearsPower(self, **kwargs) -> float:
        return self._ti.i_bears_power(**kwargs)

    def iBands(self, **kwargs) -> float:
        return self._ti.i_bands(**kwargs)

    def iBandsOnArray(self, **kwargs) -> float:
        return self._ti.i_bands_on_array(**kwargs)

    def iBullsPower(self, **kwargs) -> float:
        return self._ti.i_bulls_power(**kwargs)

    def iCCI(self, **kwargs) -> float:
        return self._ti.i_cci(**kwargs)

    def iCCIOnArray(self, **kwargs) -> float:
        return self._ti.i_cci_on_array(**kwargs)

    # Only support string input for the custom inputs
    def iCustom(self, **kwargs) -> float:
        return self._ti.i_custom(**kwargs)

    def iDeMarker(self, **kwargs) -> float:
        return self._ti.i_de_marker(**kwargs)

    def iEnvelopes(self, **kwargs) -> float:
        return self._ti.i_envelopes(**kwargs)

    def iEnvelopesOnArray(self, **kwargs) -> float:
        return self._ti.i_envelopes_on_array(**kwargs)

    def iForce(self, **kwargs) -> float:
        return self._ti.i_force(**kwargs)

    def iFractals(self, **kwargs) -> float:
        return self._ti.i_fractals(**kwargs)

    def iGator(self, **kwargs) -> float:
        return self._ti.i_gator(**kwargs)

    def iIchimoku(self, **kwargs) -> float:
        return self._ti.i_ichimoku(**kwargs)

    def iBWMFI(self, **kwargs) -> float:
        return self._ti.i_bwmfi(**kwargs)

    def iMomentum(self, **kwargs) -> float:
        return self._ti.i_momentum(**kwargs)

    def iMomentumOnArray(self, **kwargs) -> float:
        return self._ti.i_momentum_on_array(**kwargs)

    def iMFI(self, **kwargs) -> float:
        return self._ti.i_mfi(**kwargs)

    def iMA(self, **kwargs) -> float:
        return self._ti.i_ma(**kwargs)

    def iMAOnArray(self, **kwargs) -> float:
        return self._ti.i_ma_on_array(**kwargs)

    def iOsMA(self, **kwargs) -> float:
        return self._ti.i_os_ma(**kwargs)

    def iMACD(self, **kwargs) -> float:
        return self._ti.i_macd(**kwargs)

    def iOBV(self, **kwargs) -> float:
        return self._ti.i_obv(**kwargs)

    def iSAR(self, **kwargs) -> float:
        return self._ti.i_sar(**kwargs)

    def iRSI(self, **kwargs) -> float:
        return self._ti.i_rsi(**kwargs)

    def iRSIOnArray(self, **kwargs) -> float:
        return self._ti.i_rsi_on_array(**kwargs)

    def iRVI(self, **kwargs) -> float:
        return self._ti.i_rvi(**kwargs)

    def iStdDev(self, **kwargs) -> float:
        return self._ti.i_std_dev(**kwargs)

    def iStdDevOnArray(self, **kwargs) -> float:
        return self._ti.i_std_dev_on_array(**kwargs)

    def iStochastic(self, **kwargs) -> float:
        return self._ti.i_stochastic(**kwargs)

    def iWPR(self, **kwargs) -> float:
        return self._ti.i_wpr(**kwargs)

    # mt4 checkup functions
    def GetLastError(self) -> int:
        return self._cf.get_int_from_mt(fun_name="GetLastError")

    # mt4 account information
    def AccountInfoDouble(self, **kwargs) -> float:
        return self._ai.account_info_double(**kwargs)

    def AccountInfoInteger(self, **kwargs) -> int:
        return self._ai.account_info_integer(**kwargs)

    def AccountInfoString(self, **kwargs) -> str:
        return self._ai.account_info_string(**kwargs)

    def AccountBalance(self) -> float:
        return self._cf.get_float_from_mt(fun_name="AccountBalance")

    def AccountCredit(self) -> float:
        return self._cf.get_float_from_mt(fun_name="AccountCredit")

    def AccountCompany(self) -> str:
        return self._cf.get_str_from_mt(fun_name="AccountCompany")

    def AccountCurrency(self) -> str:
        return self._cf.get_str_from_mt(fun_name="AccountCurrency")

    def AccountEquity(self) -> float:
        return self._cf.get_float_from_mt(fun_name="AccountEquity")

    def AccountFreeMargin(self) -> float:
        return self._cf.get_float_from_mt(fun_name="AccountFreeMargin")

    def AccountFreeMarginCheck(self, **kwargs) -> float:
        return self._ai.account_free_margin_check(**kwargs)

    def AccountFreeMarginMode(self) -> float:
        return self._cf.get_float_from_mt(fun_name="AccountFreeMarginMode")

    def AccountLeverage(self) -> int:
        return self._cf.get_int_from_mt(fun_name="AccountLeverage")

    def AccountMargin(self) -> float:
        return self._cf.get_float_from_mt(fun_name="AccountMargin")

    def AccountName(self) -> str:
        return self._cf.get_str_from_mt(fun_name="AccoauntName")

    def AccountNumber(self) -> int:
        return self._cf.get_int_from_mt(fun_name="AccountNumber")

    def AccountProfit(self) -> float:
        return self._cf.get_float_from_mt(fun_name="AccountProfit")

    def AccountServer(self) -> str:
        return self._cf.get_str_from_mt(fun_name="AccountServer")

    def AccountStopoutLevel(self) -> int:
        return self._cf.get_int_from_mt(fun_name="AccountStopoutLevel")

    def AccountStopoutMode(self) -> int:
        return self._cf.get_int_from_mt(fun_name="AccountStopoutMode")

    # market info
    def MarketInfo(self, **kwargs) -> float:
        return self._mi.market_info(**kwargs)

    def SymbolsTotal(self, **kwargs) -> int:
        return self._mi.symbols_total(**kwargs)

    def SymbolName(self, **kwargs) -> str:
        return self._mi.symbol_name(**kwargs)

    def SymbolSelect(self, **kwargs) -> bool:
        return self._mi.symbol_select(**kwargs)

    def SymbolInfoDouble(self, **kwargs) -> float:
        return self._mi.symbol_info_double(**kwargs)

    def SymbolInfoInteger(self, **kwargs) -> int:
        return self._mi.symbol_info_integer(**kwargs)

    def SymbolInfoString(self, **kwargs) -> str:
        return self._mi.symbol_info_string(**kwargs)

    def SymbolInfoTick(self, **kwargs) -> Union[datetime, float, float, float, int]:
        return self._mi.symbol_info_tick(**kwargs)

    def SymbolInfoSessionQuote(self, **kwargs) -> Union[datetime, datetime]:
        return self._mi.symbol_info_session_quote(**kwargs)

    def SymbolInfoSessionTrade(self, **kwargs) -> Union[datetime, datetime]:
        return self._mi.symbol_info_session_trade(**kwargs)

    # timeseries and indicators access
    def SeriesInfoInteger(self, **kwargs) -> int:
        return self._taia.series_info_integer(**kwargs)

    def RefreshRates(self, **kwargs) -> bool:
        return self._cf.get_bool_from_mt(fun_name="RefreshRates")

    def CopyRates_001(self, **kwargs) -> List[Union[datetime, float, float, float, float, int, int, int]]:
        return self._taia.copy_rates_001(**kwargs)

    def CopyRates_002(self, **kwargs) -> List[Union[datetime, float, float, float, float, int, int, int]]:
        return self._taia.copy_rates_002(**kwargs)

    def CopyRates_003(self, **kwargs) -> List[Union[datetime, float, float, float, float, int, int, int]]:
        return self._taia.copy_rates_003(**kwargs)

    def CopyTime_001(self, **kwargs) -> List[datetime]:
        return self._taia.copy_time_001(**kwargs)

    def CopyTime_002(self, **kwargs) -> List[datetime]:
        return self._taia.copy_time_002(**kwargs)

    def CopyTime_003(self, **kwargs) -> List[datetime]:
        return self._taia.copy_time_003(**kwargs)

    def CopyOpen_001(self, **kwargs) -> List[float]:
        return self._taia.copy_open_001(**kwargs)

    def CopyOpen_002(self, **kwargs) -> List[float]:
        return self._taia.copy_open_002(**kwargs)

    def CopyOpen_003(self, **kwargs) -> List[float]:
        return self._taia.copy_open_003(**kwargs)

    def CopyHigh_001(self, **kwargs) -> List[float]:
        return self._taia.copy_high_001(**kwargs)

    def CopyHigh_002(self, **kwargs) -> List[float]:
        return self._taia.copy_high_002(**kwargs)

    def CopyHigh_003(self, **kwargs) -> List[float]:
        return self._taia.copy_high_003(**kwargs)

    def CopyLow_001(self, **kwargs) -> List[float]:
        return self._taia.copy_low_001(**kwargs)

    def CopyLow_002(self, **kwargs) -> List[float]:
        return self._taia.copy_low_002(**kwargs)

    def CopyLow_003(self, **kwargs) -> List[float]:
        return self._taia.copy_low_003(**kwargs)

    def CopyClose_001(self, **kwargs) -> List[float]:
        return self._taia.copy_close_001(**kwargs)

    def CopyClose_002(self, **kwargs) -> List[float]:
        return self._taia.copy_close_002(**kwargs)

    def CopyClose_003(self, **kwargs) -> List[float]:
        return self._taia.copy_close_003(**kwargs)

    def CopyTickVolume_001(self, **kwargs) -> List[int]:
        return self._taia.copy_tick_volume_001(**kwargs)

    def CopyTickVolume_002(self, **kwargs) -> List[int]:
        return self._taia.copy_tick_volume_002(**kwargs)

    def CopyTickVolume_003(self, **kwargs) -> List[int]:
        return self._taia.copy_tick_volume_003(**kwargs)

    def Bars_001(self, **kwargs) -> int:
        return self._taia.bars_001(**kwargs)

    def Bars_002(self, **kwargs) -> int:
        return self._taia.bars_002(**kwargs)

    def iBars(self, **kwargs) -> int:
        return self._taia.ibars(**kwargs)

    def iBarShift(self, **kwargs) -> int:
        return self._taia.ibar_shift(**kwargs)

    def iClose(self, **kwargs) -> float:
        return self._taia.iclose(**kwargs)

    def iHigh(self, **kwargs) -> float:
        return self._taia.ihigh(**kwargs)

    def iHighest(self, **kwargs) -> float:
        return self._taia.ihighest(**kwargs)

    def iLow(self, **kwargs) -> float:
        return self._taia.ilow(**kwargs)

    def iLowest(self, **kwargs) -> float:
        return self._taia.ilowest(**kwargs)

    def iOpen(self, **kwargs) -> float:
        return self._taia.iopen(**kwargs)

    def iTime(self, **kwargs) -> datetime:
        return self._taia.itime(**kwargs)

    def iVolume(self, **kwargs) -> int:
        return self._taia.ivolume(**kwargs)

    # mt4 conversion functions
    def NormalizeDouble(self, value: float, digits: int):
        # https://stackoverflow.com/a/41383900/2358836
        factor = 1 / (10 ** digits)
        return (value // factor) * factor

    def __UpdatePredefiniedVariables(self, parts: List[str]) -> None:
        self._pv.digits = int(parts[0])
        self._pv.point = float(parts[1])
        self._pv.last_error = int(parts[2])
        self._pv.period = int(parts[3])
        self._pv.symbol = parts[4]
        self._pv.ask = float(parts[5])
        self._pv.bars = int(parts[6])
        self._pv.bid = float(parts[7])
        self._pv.is_connected = self.__GetBoolFromString(data=parts[8])
        self._pv.is_optimization = self.__GetBoolFromString(data=parts[9])
        self._pv.is_testing = self.__GetBoolFromString(data=parts[10])
        self._pv.is_trade_allowed = self.__GetBoolFromString(data=parts[11])

        end_pos: int = 12
        self._pv.close, end_pos = self.__GetFloatList(parts=parts, start_idx=end_pos)
        self._pv.high, end_pos = self.__GetFloatList(parts=parts, start_idx=end_pos)
        self._pv.low, end_pos = self.__GetFloatList(parts=parts, start_idx=end_pos)
        self._pv.open, end_pos = self.__GetFloatList(parts=parts, start_idx=end_pos)
        self._pv.time, end_pos = self.__GetTimeList(parts=parts, start_idx=end_pos)
        self._pv.volume, end_pos = self.__GetIntList(parts=parts, start_idx=end_pos)

    def __GetBoolFromString(self, data: str) -> bool:
        if data == "0":
            return False
        else:
            return True

    def __GetTimeList(self, parts: List[str], start_idx: int) -> Union[List[datetime], int]:
        result: List[datetime] = list()

        data_len: int = int(parts[start_idx])
        start_idx += 1
        end_pos: int = start_idx + data_len

        i: int = start_idx
        while i < end_pos:
            result.append(datetime.strptime(parts[i], "%Y.%m.%d %H:%M:%S"))
            i += 1

        return (result, end_pos)

    def __GetIntList(self, parts: List[str], start_idx: int) -> Union[List[int], int]:
        result: List[int] = list()

        data_len: int = int(parts[start_idx])
        start_idx += 1
        end_pos: int = start_idx + data_len

        i: int = start_idx
        while i < end_pos:
            result.append(int(parts[i]))
            i += 1

        return (result, end_pos)

    def __GetFloatList(self, parts: List[str], start_idx: int) -> Union[List[float], int]:
        result: List[float] = list()

        data_len: int = int(parts[start_idx])
        start_idx += 1
        end_pos: int = start_idx + data_len

        i: int = start_idx
        while i < end_pos:
            result.append(float(parts[i]))
            i += 1

        return (result, end_pos)

    def __create_fun_map(self) -> Dict[str, Any]:
        fun_map = dict()

        # basic callbacks
        fun_map["OnInit"] = self.OnInit
        fun_map["OnTick"] = self.OnTick
        fun_map["OnTester"] = self.OnTester
        fun_map["OnChartEvent"] = self.OnChartEvent
        fun_map["OnTimer"] = self.OnTimer
        fun_map["OnDeinit"] = self.OnDeinit

        return fun_map



from typing import Dict, List, Set, Tuple, Union, Any

import pysailfish.internal.MT_EA.mt4_const as mc
from pysailfish.internal.DataCtrl.MT4DataCtrl.MT4DataCtrl import MT4DataCtrl

class common_functions(object):
    def __init__(self):
        self.__data_ctrl: MT4DataCtrl = None
        self.__fun_num_map: Dict[str, int] = dict()

    def init_component(self, data_ctrl: MT4DataCtrl, fun_num_map: Dict[str, int]):
        self.__data_ctrl = data_ctrl
        self.__fun_num_map = fun_num_map

    def get_datetime_from_mt(self, fun_name: str) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map[fun_name]))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_datetime_from_mt4()

    def get_int_from_mt(self, fun_name: str) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map[fun_name]))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_int_from_mt4()

    def get_float_from_mt(self, fun_name: str) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map[fun_name]))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def get_str_from_mt(self, fun_name: str) -> str:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map[fun_name]))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_string_from_mt4()

    def get_bool_from_mt(self, fun_name: str) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map[fun_name]))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()



from typing import Dict, List, Set, Tuple, Union, Any
from datetime import datetime

import pysailfish.internal.MT_EA.mt4_const as mc
from pysailfish.internal.DataCtrl.MT4DataCtrl.MT4DataCtrl import MT4DataCtrl

class trade_functions(object):
    def __init__(self):
        self.__data_ctrl: MT4DataCtrl = None
        self.__fun_num_map: Dict[str, int] = dict()

    def init_component(self, data_ctrl: MT4DataCtrl, fun_num_map: Dict[str, int]):
        self.__data_ctrl = data_ctrl
        self.__fun_num_map = fun_num_map

    def order_close(self, ticket: int, lots: float, price: float, slippage: int, arrow_color: str) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["OrderClose"]))
        self.__data_ctrl.add_data_to_msg(data=str(ticket))
        self.__data_ctrl.add_data_to_msg(data=str(lots))
        self.__data_ctrl.add_data_to_msg(data=str(price))
        self.__data_ctrl.add_data_to_msg(data=str(slippage))
        self.__data_ctrl.add_data_to_msg(data=arrow_color)
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    def order_close_by(self, ticket: int, opposite: int, arrow_color: str) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["OrderCloseBy"]))
        self.__data_ctrl.add_data_to_msg(data=str(ticket))
        self.__data_ctrl.add_data_to_msg(data=str(opposite))
        self.__data_ctrl.add_data_to_msg(data=arrow_color)
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    def order_delete(self, ticket: int, arrow_color: str) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["OrderDelete"]))
        self.__data_ctrl.add_data_to_msg(data=str(ticket))
        self.__data_ctrl.add_data_to_msg(data=arrow_color)
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    def order_modify(self, ticket: int, price: float, stoploss: float, takeprofit: float, expiration: datetime, arrow_color: str) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["OrderModify"]))
        self.__data_ctrl.add_data_to_msg(data=str(ticket))
        self.__data_ctrl.add_data_to_msg(data=str(price))
        self.__data_ctrl.add_data_to_msg(data=str(stoploss))
        self.__data_ctrl.add_data_to_msg(data=str(takeprofit))
        self.__data_ctrl.add_data_to_msg(data=expiration.strftime("%Y.%m.%d %H:%M:%S"))
        self.__data_ctrl.add_data_to_msg(data=arrow_color)
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    def order_print(self) -> None:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["OrderPrint"]))

    def order_select(self, index: int, select: int, pool: int = mc.MODE_TRADES) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["OrderSelect"]))
        self.__data_ctrl.add_data_to_msg(data=str(index))
        self.__data_ctrl.add_data_to_msg(data=str(select))
        self.__data_ctrl.add_data_to_msg(data=str(pool))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    def order_send(self
                   , symbol: str
                   , cmd: int
                   , volume: float
                   , price: float
                   , slippage: int
                   , stoploss: float
                   , takeprofit: float
                   , comment: str = ""
                   , magic: int = 0
                   , expiration: datetime = datetime(1970, 1, 1, 0, 0, 0)
                   , arrow_color: str = mc.clrNONE):
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["OrderSend"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(cmd))
        self.__data_ctrl.add_data_to_msg(data=str(volume))
        self.__data_ctrl.add_data_to_msg(data=str(price))
        self.__data_ctrl.add_data_to_msg(data=str(slippage))
        self.__data_ctrl.add_data_to_msg(data=str(stoploss))
        self.__data_ctrl.add_data_to_msg(data=str(takeprofit))
        self.__data_ctrl.add_data_to_msg(data=comment)
        self.__data_ctrl.add_data_to_msg(data=str(magic))
        self.__data_ctrl.add_data_to_msg(data=expiration.strftime("%Y.%m.%d %H:%M:%S"))
        self.__data_ctrl.add_data_to_msg(data=arrow_color)
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_int_from_mt4()

    def order_close(self
                    , ticket: int
                    , lots: float
                    , price: float
                    , slippage: int
                    , arrow_color: str) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["OrderClose"]))
        self.__data_ctrl.add_data_to_msg(data=str(ticket))
        self.__data_ctrl.add_data_to_msg(data=str(lots))
        self.__data_ctrl.add_data_to_msg(data=str(price))
        self.__data_ctrl.add_data_to_msg(data=str(slippage))
        self.__data_ctrl.add_data_to_msg(data=arrow_color)
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()



from datetime import datetime

import pysailfish.internal.MT_EA.mt4_const as mc
from pysailfish.internal.MT_EA.MT4_EA import MT4_EA

class MA_EA(MT4_EA):
    def __init__(self):
        super().__init__()

    # override
    def _OnInit(self) -> int:
        self._logger.info(self._user_inputs)
        self._logger.info("Here")
        return 0

    # override
    def _OnDeinit(self, reason: int) -> None:
        self._logger.info(f"Here reason: {reason}")
        return None

    # override
    def _OnTick(self) -> None:
        vv = self.iADX(symbol=self._pv.symbol
                       , timeframe=mc.PERIOD_CURRENT
                       , period=self._user_inputs["MovingPeriod"]
                       , applied_price=mc.PRICE_CLOSE
                       , mode=mc.MODE_MAIN
                       , shift=0)
        # self._logger.info(f"Here {self._pv.symbol} {self._pv.time[0]}")
        # --- check for history and trading
        if self._pv.bars < 100 or self._pv.is_trade_allowed == False:
            return

        # --- calculate open orders by current symbol
        if self.__calculate_current_orders(self._pv.symbol) == 0:
            self.__check_for_open()
        else:
            self.__check_for_close()
        return None

    # override
    def _OnTimer(self) -> None:
        return None

    # override
    def _OnTester(self) -> float:
        self._logger.info("Here")
        return 0.0

    # override
    def _OnChartEvent(self
                      , id: int
                      , lparam: int
                      , dparam: float
                      , sparam: str) -> None:
        self._logger.info(f"Here id: {id} lparam: {lparam} dparam: {dparam} sparam: {sparam}")
        return None

    def __check_for_close(self) -> None:
        ma: float = 0
        #--- go trading only for first tiks of new bar
        if self._pv.volume[0] > 1:
            return None
        #--- get Moving Average
        ma = self.iMA(symbol=self._pv.symbol
                      , timeframe=mc.PERIOD_CURRENT
                      , ma_period=self._user_inputs["MovingPeriod"]
                      , ma_shift=self._user_inputs["MovingShift"]
                      , ma_method=mc.MODE_SMA
                      , applied_price=mc.PRICE_CLOSE
                      , shift=0)
        #---
        for i in range(self.OrdersTotal()):
            if self.OrderSelect(index=i, select=mc.SELECT_BY_POS, pool=mc.MODE_TRADES) == False:
                break
            if self.OrderMagicNumber() != self._pv.magic_num or self.OrderSymbol() != self._pv.symbol:
                continue
            #--- check order type
            if self.OrderType() == mc.OP_BUY:
                if self._pv.open[1] > ma and self._pv.close[1] < ma:
                    if not self.OrderClose(ticket=self.OrderTicket(), lots=self.OrderLots(), price=self._pv.bid, slippage=3, arrow_color=mc.clrWhite):
                        self._logger.error(f"OrderClose error {self.GetLastError()}")
                break
            if self.OrderType() == mc.OP_SELL:
                if self._pv.open[1] < ma and self._pv.close[1] > ma:
                    if not self.OrderClose(ticket=self.OrderTicket(), lots=self.OrderLots(), price=self._pv.ask, slippage=3, arrow_color=mc.clrWhite):
                        self._logger.error(f"OrderClose error {self.GetLastError()}");
                break

    def __lots_optimized(self) -> float:
        lot: float = float(self._user_inputs["Lots"])
        maximum_risk: float = float(self._user_inputs["MaximumRisk"])
        decrease_factor: float = float(self._user_inputs["DecreaseFactor"])
        orders: int = self.OrdersHistoryTotal() # history orders total
        losses: int = 0 # number of losses orders without a break
        #--- select lot size
        lot = self.NormalizeDouble(value=(self.AccountFreeMargin() * maximum_risk / 1000.0), digits=1);
        #--- calcuulate number of losses orders without a break
        if decrease_factor > 0:
            for i in reversed(range(orders)):
                if self.OrderSelect(index=i, select=mc.SELECT_BY_POS, pool=mc.MODE_HISTORY) == False:
                    self._logger.error("Error in history")
                    break
                if self.OrderSymbol() != self._pv.symbol or self.OrderType() > mc.OP_SELL:
                    continue
                #---
                if self.OrderProfit() > 0:
                    break
                if self.OrderProfit() < 0:
                    losses += 1
            if losses > 1:
                lot = self.NormalizeDouble(value=(lot - lot * losses / decrease_factor), digits=1)
        #--- return lot size
        if lot < 0.1:
            lot = 0.1
        return lot

    def __check_for_open(self) -> None:
        ma: float = 0
        res: int = 0
        #--- go trading only for first tiks of new bar
        if self._pv.volume[0] > 1:
            return None
        #--- get Moving Average
        ma = self.iMA(symbol=self._pv.symbol
                      , timeframe=mc.PERIOD_CURRENT
                      , ma_period=self._user_inputs["MovingPeriod"]
                      , ma_shift=self._user_inputs["MovingShift"]
                      , ma_method=mc.MODE_SMA
                      , applied_price=mc.PRICE_CLOSE
                      , shift=0)
        #--- sell conditions
        if self._pv.open[1] > ma and self._pv.close[1] < ma:
            res = self.OrderSend(symbol=self._pv.symbol
                                 , cmd=mc.OP_SELL
                                 , volume=self.__lots_optimized()
                                 , price=self._pv.bid
                                 , slippage=3
                                 , stoploss=0
                                 , takeprofit=0
                                 , comment=""
                                 , magic=self._pv.magic_num
                                 , expiration=datetime(1970, 1, 1, 0, 0, 0)
                                 , arrow_color=mc.clrRed)
            return None
        #--- buy conditions
        if self._pv.open[1] < ma and self._pv.close[1] > ma:
            res = self.OrderSend(symbol=self._pv.symbol
                                 , cmd=mc.OP_BUY
                                 , volume=self.__lots_optimized()
                                 , price=self._pv.ask
                                 , slippage=3
                                 , stoploss=0
                                 , takeprofit=0
                                 , comment=""
                                 , magic=self._pv.magic_num
                                 , expiration=datetime(1970, 1, 1, 0, 0, 0)
                                 , arrow_color=mc.clrBlue)
            return None

    def __calculate_current_orders(self, symbol: str) -> int:
        buys: int = 0
        sells: int = 0
        # ---
        for i in range(self.OrdersTotal()):
            if self._tf.order_select(index=i, select=mc.SELECT_BY_POS, pool=mc.MODE_TRADES) == False:
                break
            if self.OrderSymbol() == self._pv.symbol and self.OrderMagicNumber() == self._pv.magic_num:
                if self.OrderType() == mc.OP_BUY:
                    buys += 1
                if self.OrderType() == mc.OP_SELL:
                    sells += 1
        # --- return orders volume
        if buys > 0:
            return buys
        else:
            return -sells

def main() -> None:
    ea = MA_EA()
    ea.InitComponent(server_ip="127.0.0.1"
                     , server_port=23456
                     , ea_name="MA_EA")
    ea.StartEA()

if __name__ == "__main__":
    main()



from pysailfish.internal.MT_EA.MT4_EA import MT4_EA

class MT4DemoEA(MT4_EA):
    def __init__(self):
        super().__init__()

    # override
    def _OnInit(self) -> int:
        self._logger.info(self._user_inputs)
        self._logger.info("Here")
        return 0

    # override
    def _OnDeinit(self, reason: int) -> None:
        self._logger.info(f"Here reason: {reason}")
        return None

    # override
    def _OnTick(self) -> None:
        self._logger.info("Here")
        return None

    # override
    def _OnTimer(self) -> None:
        return None

    # override
    def _OnTester(self) -> float:
        self._logger.info("Here")
        return 0.0

    # override
    def _OnChartEvent(self
                      , id: int
                      , lparam: int
                      , dparam: float
                      , sparam: str) -> None:
        self._logger.info(f"Here id: {id} lparam: {lparam} dparam: {dparam} sparam: {sparam}")
        return None

def main() -> None:
    ea = MT4DemoEA()
    ea.InitComponent(server_ip="127.0.0.1"
                     , server_port=23456
                     , ea_name="MT4DemoEA")
    ea.StartEA()

if __name__ == "__main__":
    main()



mt4_fun_num_map = {
"CopyLow_001": 1
, "TimeHour": 2
, "StringToInteger": 3
, "AccountEquity": 4
, "GlobalVariableDel": 5
, "IsTradeAllowed": 6
, "FolderDelete": 7
, "MQLInfoInteger": 8
, "SignalInfoGetString": 9
, "OrdersHistoryTotal": 10
, "ObjectsDeleteAll_001": 11
, "CharArrayToString": 12
, "StringToDouble": 13
, "iBarShift": 14
, "ObjectName": 15
, "StringTrimRight": 16
, "ArrayFree": 17
, "WindowFirstVisibleBar": 18
, "StringSubstr": 19
, "iAD": 20
, "SignalBaseTotal": 21
, "ObjectGetShiftByValue": 22
, "StringGetCharacter": 23
, "fmax": 24
, "SetIndexEmptyValue": 25
, "ChartScreenShot": 26
, "StringLen": 27
, "OrderSend": 28
, "OrderDelete": 29
, "OrderProfit": 30
, "StringSetCharacter": 31
, "WindowRedraw": 32
, "AccountName": 33
, "CryptDecode": 34
, "FolderClean": 35
, "IsTesting": 36
, "ChartFirst": 37
, "OrderModify": 38
, "HideTestIndicators": 39
, "TesterStatistics": 40
, "ExpertRemove": 41
, "ceil": 42
, "SignalBaseGetInteger": 43
, "AccountCurrency": 44
, "ChartIndicatorsTotal": 45
, "StringAdd": 46
, "ObjectGetDouble": 47
, "DayOfYear": 48
, "iMA": 49
, "MathRand": 50
, "iAO": 51
, "iSAR": 52
, "cos": 53
, "iOpen": 54
, "Symbol": 55
, "WindowExpertName": 56
, "Year": 57
, "GlobalVariableGet": 58
, "ObjectSetString_002": 59
, "iLow": 60
, "DebugBreak": 61
, "WebRequest": 62
, "StringInit": 63
, "GetLastError": 64
, "OnInit": 65
, "FolderCreate": 66
, "WindowOnDropped": 67
, "GlobalVariableTime": 68
, "FileMove": 69
, "FileOpen": 70
, "FileWriteString": 71
, "iTime": 72
, "CheckPointer": 73
, "OrderSwap": 74
, "ChartSaveTemplate": 75
, "TextOut": 76
, "Sleep": 77
, "MathCos": 78
, "sin": 79
, "ChartID": 80
, "CopyTickVolume_001": 81
, "StringCompare": 82
, "ObjectSetFiboDescription": 83
, "TimeMinute": 84
, "FileOpenHistory": 85
, "ZeroMemory": 86
, "iCustom": 87
, "srand": 88
, "SendFTP": 89
, "SignalInfoGetDouble": 90
, "AccountNumber": 91
, "StringFind": 92
, "IsLibrariesAllowed": 93
, "iEnvelopes": 94
, "StringFormat": 95
, "ArrayCopy": 96
, "iWPR": 97
, "WindowYOnDropped": 98
, "ArrayMinimum": 99
, "OrderSymbol": 100
, "CopyTime_003": 101
, "StringFill": 102
, "DayOfWeek": 103
, "AccountStopoutMode": 104
, "Day": 105
, "GetTickCount": 106
, "ArrayCompare": 107
, "ObjectGetTimeByValue": 108
, "SignalInfoSetDouble": 109
, "TimeDayOfYear": 110
, "GlobalVariableTemp": 111
, "FileFlush": 112
, "StringSplit": 113
, "rand": 114
, "IndicatorSetString": 115
, "MathArctan": 116
, "GetPointer": 117
, "AccountProfit": 118
, "Hour": 119
, "IndicatorSetDouble": 120
, "MathMax": 121
, "iRVI": 122
, "Digits": 123
, "FileDelete": 124
, "MathCeil": 125
, "OnChartEvent": 126
, "CopyRates_003": 127
, "StringReplace": 128
, "iAlligator": 129
, "FileWriteLong": 130
, "WindowPriceMax": 131
, "iRSIOnArray": 132
, "iHighest": 133
, "round": 134
, "CopyRates_002": 135
, "FileTell": 136
, "Bars_002": 137
, "TimeCurrent": 138
, "TerminalInfoInteger": 139
, "CopyTickVolume_003": 140
, "MathArccos": 141
, "TextGetSize": 142
, "AccountMargin": 143
, "MarketInfo": 144
, "WindowPriceOnDropped": 145
, "ObjectsDeleteAll_003": 146
, "OnDeinit": 147
, "ChartGetInteger": 148
, "GlobalVariablesFlush": 149
, "ObjectFind_002": 150
, "ResourceCreate": 151
, "GlobalVariableName": 152
, "FileReadDatetime": 153
, "ObjectGetValueByShift": 154
, "SetIndexLabel": 155
, "Minute": 156
, "WindowXOnDropped": 157
, "SetIndexArrow": 158
, "FileReadLong": 159
, "fmin": 160
, "IsExpertEnabled": 161
, "Point": 162
, "ObjectGetString": 163
, "EnumToString": 164
, "MathIsValidNumber": 165
, "ColorToARGB": 166
, "FileFindClose": 167
, "StringToColor": 168
, "StringTrimLeft": 169
, "iIchimoku": 170
, "TerminalCompany": 171
, "NormalizeDouble": 172
, "ChartWindowOnDropped": 173
, "GlobalVariablesDeleteAll": 174
, "ChartWindowFind": 175
, "MathExp": 176
, "StringToShortArray": 177
, "OrdersTotal": 178
, "TimeDayOfWeek": 179
, "IsDemo": 180
, "MQLInfoString": 181
, "CopyLow_002": 182
, "SymbolInfoDouble": 183
, "EventSetMillisecondTimer": 184
, "AccountBalance": 185
, "atan": 186
, "WindowBarsPerChart": 187
, "SetIndexBuffer": 188
, "MathLog10": 189
, "ObjectDelete_002": 190
, "AccountInfoInteger": 191
, "OrderOpenPrice": 192
, "GlobalVariableSet": 193
, "CopyHigh_003": 194
, "StringBufferLen": 195
, "MathRound": 196
, "FileFindFirst": 197
, "WindowsTotal": 198
, "FileIsExist": 199
, "ArrayMaximum": 200
, "iAC": 201
, "StrToDouble": 202
, "ChartTimeOnDropped": 203
, "ArrayGetAsSeries": 204
, "ArrayIsDynamic": 205
, "MathArcsin": 206
, "SymbolInfoTick": 207
, "TimeDaylightSavings": 208
, "IntegerToString": 209
, "AccountFreeMargin": 210
, "SymbolInfoSessionQuote": 211
, "fabs": 212
, "EventKillTimer": 213
, "ChartSetDouble": 214
, "iBullsPower": 215
, "fmod": 216
, "OrderPrint": 217
, "SetIndexShift": 218
, "PrintFormat": 219
, "MathSqrt": 220
, "ShortToString": 221
, "FileIsEnding": 222
, "PeriodSeconds": 223
, "IsStopped": 224
, "OnTimer": 225
, "pow": 226
, "FileWrite": 227
, "ResourceFree": 228
, "CopyHigh_001": 229
, "iStochastic": 230
, "OrderOpenTime": 231
, "DoubleToString": 232
, "GlobalVariableSetOnCondition": 233
, "ChartTimePriceToXY": 234
, "ArrayDimension": 235
, "ChartIndicatorDelete": 236
, "OrderExpiration": 237
, "SendMail": 238
, "ChartSetSymbolPeriod": 239
, "ArrayCopyRates": 240
, "SeriesInfoInteger": 241
, "MQLSetInteger": 242
, "iATR": 243
, "FileReadArray": 244
, "ObjectGetValueByTime": 245
, "CharToString": 246
, "ObjectDescription": 247
, "FileReadDouble": 248
, "FileWriteStruct": 249
, "FileFindNext": 250
, "OrderCloseTime": 251
, "iMACD": 252
, "iOsMA": 253
, "floor": 254
, "StrToTime": 255
, "RefreshRates": 256
, "iBands": 257
, "TerminalClose": 258
, "SignalInfoGetInteger": 259
, "ResourceSave": 260
, "iLowest": 261
, "CopyLow_003": 262
, "FileWriteFloat": 263
, "ResourceReadImage": 264
, "MathSin": 265
, "ChartPriceOnDropped": 266
, "FileReadBool": 267
, "ShortArrayToString": 268
, "ChartIndicatorName": 269
, "SetIndexDrawBegin": 270
, "ObjectGetInteger": 271
, "SignalUnsubscribe": 272
, "IndicatorBuffers": 273
, "EventChartCustom": 274
, "iBandsOnArray": 275
, "FileCopy": 276
, "OrderClose": 277
, "ChartOpen": 278
, "iMFI": 279
, "ObjectsTotal_001": 280
, "OnTester": 281
, "CopyClose_001": 282
, "FileGetInteger": 283
, "asin": 284
, "FileReadNumber": 285
, "AccountInfoDouble": 286
, "CopyTime_001": 287
, "CopyClose_002": 288
, "SendNotification": 289
, "acos": 290
, "EventSetTimer": 291
, "CopyTickVolume_002": 292
, "IndicatorSetInteger": 293
, "SignalInfoSetInteger": 294
, "MathLog": 295
, "ObjectSetInteger_002": 296
, "CopyRates_001": 297
, "log10": 298
, "ChartApplyTemplate": 299
, "ObjectCreate_002": 300
, "MathMin": 301
, "OrderType": 302
, "ObjectType": 303
, "ArrayResize": 304
, "MathTan": 305
, "CopyOpen_003": 306
, "SymbolName": 307
, "CopyOpen_001": 308
, "MathSrand": 309
, "FileSize": 310
, "OrderMagicNumber": 311
, "ChartXYToTimePrice": 312
, "MathFloor": 313
, "MessageBox": 314
, "FileWriteArray": 315
, "GlobalVariablesTotal": 316
, "iGator": 317
, "SetLevelStyle": 318
, "iStdDev": 319
, "iADX": 320
, "IndicatorCounted": 321
, "ArraySize": 322
, "FileSeek": 323
, "SignalBaseGetDouble": 324
, "iRSI": 325
, "iCCIOnArray": 326
, "iMomentumOnArray": 327
, "OrderTakeProfit": 328
, "WindowScreenShot": 329
, "ObjectSetString_001": 330
, "ChartNext": 331
, "ChartGetString": 332
, "ObjectGetFiboDescription": 333
, "exp": 334
, "ObjectSetText": 335
, "iOBV": 336
, "CopyHigh_002": 337
, "ObjectsTotal_002": 338
, "OrderLots": 339
, "iDeMarker": 340
, "ArrayRange": 341
, "DoubleToStr": 342
, "tan": 343
, "ChartRedraw": 344
, "SignalBaseSelect": 345
, "OnTick": 346
, "UninitializeReason": 347
, "log": 348
, "TextSetFont": 349
, "WindowFind": 350
, "ChartGetDouble": 351
, "SetLevelValue": 352
, "iStdDevOnArray": 353
, "ObjectsDeleteAll_002": 354
, "MathPow": 355
, "Comment": 356
, "CopyOpen_002": 357
, "SymbolsTotal": 358
, "FileReadStruct": 359
, "Month": 360
, "ArrayCopySeries": 361
, "iMAOnArray": 362
, "IsOptimization": 363
, "SymbolInfoString": 364
, "ObjectCreate_001": 365
, "TerminalInfoDouble": 366
, "OrderStopLoss": 367
, "IsConnected": 368
, "AccountFreeMarginCheck": 369
, "FileWriteInteger": 370
, "ObjectSetInteger_001": 371
, "CopyTime_002": 372
, "ArraySort": 373
, "SignalSubscribe": 374
, "AccountCredit": 375
, "OrderSelect": 376
, "FileReadFloat": 377
, "ChartNavigate": 378
, "ResetLastError": 379
, "iForce": 380
, "iBearsPower": 381
, "OrderComment": 382
, "OrderTicket": 383
, "IsVisualMode": 384
, "AccountInfoString": 385
, "SetIndexStyle": 386
, "ChartPeriod": 387
, "iVolume": 388
, "iHigh": 389
, "AccountFreeMarginMode": 390
, "Seconds": 391
, "Alert": 392
, "ArrayIsSeries": 393
, "OrderCloseBy": 394
, "StringGetChar": 395
, "ArrayFill": 396
, "StringToCharArray": 397
, "StringToLower": 398
, "Bars_001": 399
, "IndicatorShortName": 400
, "ChartSymbol": 401
, "MathMod": 402
, "ChartClose": 403
, "ChartSetString": 404
, "ObjectSet": 405
, "iBWMFI": 406
, "ChartSetInteger": 407
, "AccountCompany": 408
, "MathAbs": 409
, "ObjectGet": 410
, "FileIsLineEnding": 411
, "StringConcatenate": 412
, "TimeGMTOffset": 413
, "ObjectSetDouble_001": 414
, "AccountServer": 415
, "ArrayBsearch": 416
, "OrderClosePrice": 417
, "TerminalInfoString": 418
, "ObjectMove": 419
, "TimeToStr": 420
, "ObjectDelete_001": 421
, "FileReadInteger": 422
, "iClose": 423
, "iFractals": 424
, "StringToTime": 425
, "Print": 426
, "GlobalVariableCheck": 427
, "TimeYear": 428
, "ObjectFind_001": 429
, "OrderCommission": 430
, "SymbolSelect": 431
, "IsTradeContextBusy": 432
, "StringToUpper": 433
, "ChartXOnDropped": 434
, "iBars": 435
, "TimeToStruct": 436
, "ArraySetAsSeries": 437
, "TerminalPath": 438
, "AccountStopoutLevel": 439
, "ColorToString": 440
, "sqrt": 441
, "TimeGMT": 442
, "SignalBaseGetString": 443
, "WindowIsVisible": 444
, "WindowTimeOnDropped": 445
, "Period": 446
, "StringSetChar": 447
, "iEnvelopesOnArray": 448
, "TimeMonth": 449
, "TimeLocal": 450
, "CharToStr": 451
, "IndicatorDigits": 452
, "WindowHandle": 453
, "SymbolInfoSessionTrade": 454
, "AccountLeverage": 455
, "FileReadString": 456
, "ObjectSetDouble_002": 457
, "iCCI": 458
, "TimeSeconds": 459
, "ArrayInitialize": 460
, "CryptEncode": 461
, "StructToTime": 462
, "IsDllsAllowed": 463
, "CopyClose_003": 464
, "PlaySound": 465
, "iMomentum": 466
, "TimeDay": 467
, "FileWriteDouble": 468
, "FileClose": 469
, "TerminalName": 470
, "StrToInteger": 471
, "SymbolInfoInteger": 472
, "TimeToString": 473
, "WindowPriceMin": 474
, "ChartYOnDropped": 475
}



import logging
import time
import traceback
from typing import List, Dict, Tuple, Union, Set, Any
from datetime import datetime

import pysailfish.internal.observability.log_helper as lh
from pysailfish.internal.Network.TCPCast import TCPCast, TCPStatus
from pysailfish.internal.devTools.ThreadClass import ThreadClass

class MT4DataCtrl(ThreadClass):
    def __init__(self):
        self.__logger: logging.Logger = None
        self.__tcp_client: TCPCast = None
        self.__mt4_msg_callback: Any = None
        self.__data_pool: str = ""
        self.__cmd_start: str = "cmd_start"
        self.__cmd_end: str = "cmd_end"
        self.__cmd_end_len: int = len(self.__cmd_end)
        self.__proxy_msg: List[str] = list()

    # override
    def init_component(self
                       , logger: logging.Logger
                       , tcp_client: TCPCast
                       , mt4_msg_callback: Any):
        self.__logger = logger
        self.__tcp_client = tcp_client
        self.__mt4_msg_callback = mt4_msg_callback

        super().init_component(logger=logger)

    def is_connection_okay(self) -> bool:
        return self.__tcp_client.IsClientConnected()

    def send_message(self) -> None:
        # message start
        send_msg: str = self.__cmd_start + ","
        # message count
        send_msg += str(len(self.__proxy_msg)) + ","
        # message content
        send_msg += ",".join(self.__proxy_msg) + ","
        # message end
        send_msg += self.__cmd_end

        # send to mt4
        self.__tcp_client.ClientSend(send_msg=send_msg)

        self.__proxy_msg = list()

    def add_data_to_msg(self, data: str) -> None:
        self.__proxy_msg.append(data)

    def receive_string_from_mt4(self) -> str:
        fun_name, parts = self.receive_data_from_mt4()
        if fun_name is None:
            return ""
        else:
            return parts[0]

    def receive_bool_from_mt4(self) -> bool:
        fun_name, parts = self.receive_data_from_mt4()
        if fun_name is None or parts[0] == "":
            return False
        else:
            if parts[0] == "0":
                return False
            else:
                return True

    def receive_int_from_mt4(self) -> int:
        fun_name, parts = self.receive_data_from_mt4()
        if fun_name is None or parts[0] == "":
            return 0
        else:
            return int(parts[0])

    def receive_datetime_from_mt4(self) -> datetime:
        fun_name, parts = self.receive_data_from_mt4()
        if fun_name is None or parts[0] == "":
            return datetime(1970, 1, 1, 0, 0, 0)
        else:
            return datetime.strptime(parts[0], "%Y.%m.%d %H:%M:%S")

    def receive_float_from_mt4(self) -> float:
        fun_name, parts = self.receive_data_from_mt4()
        if fun_name is None or parts[0] == "":
            return 0
        else:
            return float(parts[0])

    def receive_data_from_mt4(self) -> Union[str, List[str]]:
        local_data_pool: str = ""
        while True:
            recv_status, recv_msg = self.__tcp_client.ClientRecv()
            if recv_status == TCPStatus.SUCCESS:
                local_data_pool += recv_msg
                if not local_data_pool.find(self.__cmd_end):
                    # command not ended then retry
                    continue

                parts: List[str] = local_data_pool.split(",")
                if len(parts) < 3:
                    self.__logger.error(f"message is not completed: {local_data_pool}")
                    return (None, None)

                cmd_count: int = int(parts[1])
                if cmd_count + 3 != len(parts):
                    # skip this command due to checksum is wrong
                    self.__logger.error(f"message checksum is wrong: {len(parts)}, expected: {cmd_count + 3}")
                    return (None, None)

                fun_name: str = parts[2]
                return (fun_name, parts[3:-1])

    # override
    def _user_main(self) -> None:
        recv_status, recv_msg = self.__tcp_client.ClientRecv()
        if recv_status == TCPStatus.SUCCESS:
            self.__data_pool += recv_msg
            if not self.__data_pool.find(self.__cmd_end):
                # command not ended
                return None

            while True:
                start_pos = -1
                end_pos = -1

                # find next start position
                start_pos = self.__data_pool.find(self.__cmd_start, start_pos + 1)
                # find next end position
                end_pos = self.__data_pool.find(self.__cmd_end, end_pos + 1)
                if start_pos == -1 or end_pos == -1:
                    # no more complete commands found
                    break
                if start_pos > end_pos:
                    # should not go to this case
                    break

                # extract one command from data pool
                # cmd: cmd_start,cmd_count,cmd_content,cmd_end
                cmd = self.__data_pool[start_pos:end_pos+self.__cmd_end_len]
                # remove old data from pool
                self.__data_pool = self.__data_pool[end_pos+self.__cmd_end_len:]

                parts: List[str] = cmd.split(",")
                if len(parts) < 3:
                    self.__logger.error(f"message is not completed: {cmd}")
                    continue

                cmd_count: int = int(parts[1])
                if cmd_count + 3 != len(parts):
                    # skip this command due to checksum is wrong
                    continue

                fun_name: str = parts[2]
                try:
                    self.__mt4_msg_callback(fun_name=fun_name, parts=parts[3:-1])
                except Exception as e:
                    self.__logger.error(e)
                    traceback.print_exc()



import json
try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode

from logging import Filter
from logging.handlers import SocketHandler

import pika
from pysailfish.internal.observability.logstash import formatter


class AMQPLogstashHandler(SocketHandler, object):
    """AMQP Log Format handler

    :param host: AMQP host (default 'localhost')
    :param port: AMQP port (default 5672)
    :param username: AMQP user name (default 'guest', which is the default for
        RabbitMQ)
    :param password: AMQP password (default 'guest', which is the default for
        RabbitMQ)

    :param exchange: AMQP exchange. Default 'logging.gelf'.
        A queue binding must be defined on the server to prevent
        log messages from being dropped.
    :param exchange_type: AMQP exchange type (default 'fanout').
    :param durable: AMQP exchange is durable (default False)
    :param virtual_host: AMQP virtual host (default '/').
    :param passive: exchange is declared passively, meaning that an error is
        raised if the exchange does not exist, and succeeds otherwise. This is
        useful if the user does not have configure permission on the exchange.

    :param tags: list of tags for a logger (default is None).
    :param message_type: The type of the message (default logstash).
    :param version: version of logstash event schema (default is 0).

    :param extra_fields: Send extra fields on the log record to graylog
        if true (the default)
    :param fqdn: Use fully qualified domain name of localhost as source
        host (socket.getfqdn()).
    :param facility: Replace facility with specified value. If specified,
        record.name will be passed as `logger` parameter.
    """

    def __init__(self, host='localhost', port=5672, username='guest',
                 password='guest', exchange='logstash', exchange_type='fanout',
                 virtual_host='/', message_type='logstash', tags=None,
                 durable=False, passive=False, version=0, extra_fields=True,
                 fqdn=False, facility=None, exchange_routing_key=''):


        # AMQP parameters
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.exchange_type = exchange_type
        self.exchange = exchange
        self.exchange_is_durable = durable
        self.declare_exchange_passively = passive
        self.virtual_host = virtual_host
        self.routing_key = exchange_routing_key

        SocketHandler.__init__(self, host, port)

        # Extract Logstash paramaters
        self.tags = tags or []
        fn = formatter.LogstashFormatterVersion1 if version == 1 \
            else formatter.LogstashFormatterVersion0
        self.formatter = fn(message_type, tags, fqdn)

        # Standard logging parameters
        self.extra_fields = extra_fields
        self.fqdn = fqdn
        self.facility = facility

    def makeSocket(self, **kwargs):

        return PikaSocket(self.host,
                          self.port,
                          self.username,
                          self.password,
                          self.virtual_host,
                          self.exchange,
                          self.routing_key,
                          self.exchange_is_durable,
                          self.declare_exchange_passively,
                          self.exchange_type)

    def makePickle(self, record):
        return self.formatter.format(record)


class PikaSocket(object):

    def __init__(self, host, port, username, password, virtual_host, exchange,
                routing_key, durable, passive, exchange_type):

        # create connection parameters
        credentials = pika.PlainCredentials(username, password)
        parameters = pika.ConnectionParameters(host, port, virtual_host,
                                               credentials)

        # create connection & channel
        self.connection = pika.BlockingConnection(parameters)
        self.channel = self.connection.channel()

        # create an exchange, if needed
        self.channel.exchange_declare(exchange=exchange,
                                      exchange_type=exchange_type,
                                      passive=passive,
                                      durable=durable)

        # needed when publishing
        self.spec = pika.spec.BasicProperties(delivery_mode=2)
        self.routing_key = routing_key
        self.exchange = exchange


    def sendall(self, data):

        self.channel.basic_publish(self.exchange,
                                   self.routing_key,
                                   data,
                                   properties=self.spec)

    def close(self):
        try:
            self.connection.close()
        except Exception:
            pass



MODE_TRADES = 0
MODE_HISTORY = 1

SELECT_BY_POS = 0
SELECT_BY_TICKET = 1

# https://docs.mql4.com/constants/tradingconstants/signalproperties
SIGNAL_BASE_BALANCE = 17
SIGNAL_BASE_EQUITY = 19
SIGNAL_BASE_GAIN = 16
SIGNAL_BASE_MAX_DRAWDOWN = 18
SIGNAL_BASE_PRICE = 15
SIGNAL_BASE_ROI = 20

SIGNAL_BASE_DATE_PUBLISHED = 1
SIGNAL_BASE_DATE_STARTED = 2
SIGNAL_BASE_ID = 0
SIGNAL_BASE_LEVERAGE = 6
SIGNAL_BASE_PIPS = 8
SIGNAL_BASE_RATING = 9
SIGNAL_BASE_SUBSCRIBERS = 4
SIGNAL_BASE_TRADES = 7
SIGNAL_BASE_TRADE_MODE = 5

SIGNAL_BASE_AUTHOR_LOGIN = 13
SIGNAL_BASE_BROKER = 10
SIGNAL_BASE_BROKER_SERVER = 11
SIGNAL_BASE_NAME = 12
SIGNAL_BASE_CURRENCY = 14

SIGNAL_INFO_EQUITY_LIMIT = 8
SIGNAL_INFO_SLIPPAGE = 9
SIGNAL_INFO_VOLUME_PERCENT = 7

SIGNAL_INFO_CONFIRMATIONS_DISABLED = 4
SIGNAL_INFO_COPY_SLTP = 3
SIGNAL_INFO_DEPOSIT_PERCENT = 5
SIGNAL_INFO_ID = 0
SIGNAL_INFO_SUBSCRIPTION_ENABLED = 1
SIGNAL_INFO_TERMS_AGREE = 2

SIGNAL_INFO_NAME = 6

# https://docs.mql4.com/constants/tradingconstants/orderproperties
OP_BUY = 0
OP_SELL = 1
OP_BUYLIMIT = 2
OP_BUYSTOP = 3
OP_SELLLIMIT = 4
OP_SELLSTOP = 5

# https://docs.mql4.com/constants/environment_state/accountinformation#enum_account_info_double
ACCOUNT_LOGIN = 0
ACCOUNT_TRADE_MODE = 32
ACCOUNT_LEVERAGE = 35
ACCOUNT_LIMIT_ORDERS = 47
ACCOUNT_MARGIN_SO_MODE = 44
ACCOUNT_TRADE_ALLOWED = 33
ACCOUNT_TRADE_EXPERT = 34

ACCOUNT_BALANCE = 37
ACCOUNT_CREDIT = 38
ACCOUNT_PROFIT = 39
ACCOUNT_EQUITY = 40
ACCOUNT_MARGIN = 41
ACCOUNT_MARGIN_FREE = 42
ACCOUNT_MARGIN_LEVEL = 43
ACCOUNT_MARGIN_SO_CALL = 45
ACCOUNT_MARGIN_SO_SO = 46
ACCOUNT_MARGIN_INITIAL = 48
ACCOUNT_MARGIN_MAINTENANCE = 49
ACCOUNT_ASSETS = 50
ACCOUNT_LIABILITIES = 51
ACCOUNT_COMMISSION_BLOCKED = 52

ACCOUNT_NAME = 1
ACCOUNT_SERVER = 3
ACCOUNT_CURRENCY = 36
ACCOUNT_COMPANY = 2

ACCOUNT_TRADE_MODE_DEMO = 0
ACCOUNT_TRADE_MODE_CONTEST = 1
ACCOUNT_TRADE_MODE_REAL = 2

ACCOUNT_STOPOUT_MODE_PERCENT = 0
ACCOUNT_STOPOUT_MODE_MONEY = 1

# https://docs.mql4.com/constants/environment_state/marketinfoconstants#enum_day_of_week
MODE_LOW = 1
MODE_HIGH = 2
MODE_TIME = 5
MODE_BID = 9
MODE_ASK = 10
MODE_POINT = 11
MODE_DIGITS = 12
MODE_SPREAD = 13
MODE_STOPLEVEL = 14
MODE_LOTSIZE = 15
MODE_TICKVALUE = 16
MODE_TICKSIZE = 17
MODE_SWAPLONG = 18
MODE_SWAPSHORT = 19
MODE_STARTING = 20
MODE_EXPIRATION = 21
MODE_TRADEALLOWED = 22
MODE_MINLOT = 23
MODE_LOTSTEP = 24
MODE_MAXLOT = 25
MODE_SWAPTYPE = 26
MODE_PROFITCALCMODE = 27
MODE_MARGINCALCMODE = 28
MODE_MARGININIT = 29
MODE_MARGINMAINTENANCE = 30
MODE_MARGINHEDGED = 31
MODE_MARGINREQUIRED = 32
MODE_FREEZELEVEL = 33
MODE_CLOSEBY_ALLOWED = 34

SYMBOL_SELECT = 0
SYMBOL_VISIBLE = 76
SYMBOL_SESSION_DEALS = 56
SYMBOL_SESSION_BUY_ORDERS = 60
SYMBOL_SESSION_SELL_ORDERS = 62
SYMBOL_VOLUME = 10
SYMBOL_VOLUMEHIGH = 11
SYMBOL_VOLUMELOW = 12
SYMBOL_TIME = 15
SYMBOL_DIGITS = 17
SYMBOL_SPREAD_FLOAT = 41
SYMBOL_SPREAD = 18
SYMBOL_TRADE_CALC_MODE = 29
SYMBOL_TRADE_MODE = 30
SYMBOL_START_TIME = 51
SYMBOL_EXPIRATION_TIME = 52
SYMBOL_TRADE_STOPS_LEVEL = 31
SYMBOL_TRADE_FREEZE_LEVEL = 32
SYMBOL_TRADE_EXEMODE = 33
SYMBOL_SWAP_MODE= 37
SYMBOL_SWAP_ROLLOVER3DAYS = 40
SYMBOL_EXPIRATION_MODE = 49
SYMBOL_FILLING_MODE = 50
SYMBOL_ORDER_MODE = 71

SYMBOL_BID = 1
SYMBOL_BIDHIGH = 2
SYMBOL_BIDLOW = 3
SYMBOL_ASK = 4
SYMBOL_ASKHIGH = 5
SYMBOL_ASKLOW = 6
SYMBOL_LAST = 7
SYMBOL_LASTHIGH = 8
SYMBOL_LASTLOW = 9
SYMBOL_POINT = 16
SYMBOL_TRADE_TICK_VALUE = 26
SYMBOL_TRADE_TICK_VALUE_PROFIT = 53
SYMBOL_TRADE_TICK_VALUE_LOSS = 54
SYMBOL_TRADE_TICK_SIZE = 27
SYMBOL_TRADE_CONTRACT_SIZE = 28
SYMBOL_VOLUME_MIN = 34
SYMBOL_VOLUME_MAX = 35
SYMBOL_VOLUME_STEP = 36
SYMBOL_VOLUME_LIMIT = 55
SYMBOL_SWAP_LONG = 38
SYMBOL_SWAP_SHORT = 39
SYMBOL_MARGIN_INITIAL = 42
SYMBOL_MARGIN_MAINTENANCE = 43
SYMBOL_MARGIN_LONG = 44
SYMBOL_MARGIN_SHORT = 45
SYMBOL_MARGIN_LIMIT = 46
SYMBOL_MARGIN_STOP = 47
SYMBOL_MARGIN_STOPLIMIT = 48
SYMBOL_SESSION_VOLUME = 57
SYMBOL_SESSION_TURNOVER = 58
SYMBOL_SESSION_INTEREST = 59
SYMBOL_SESSION_BUY_ORDERS_VOLUME = 61
SYMBOL_SESSION_SELL_ORDERS_VOLUME = 63
SYMBOL_SESSION_OPEN = 64
SYMBOL_SESSION_CLOSE = 65
SYMBOL_SESSION_AW = 66
SYMBOL_SESSION_PRICE_SETTLEMENT = 67
SYMBOL_SESSION_PRICE_LIMIT_MIN = 68
SYMBOL_SESSION_PRICE_LIMIT_MAX = 69

SYMBOL_CURRENCY_BASE = 22
SYMBOL_CURRENCY_PROFIT = 23
SYMBOL_CURRENCY_MARGIN = 24
SYMBOL_DESCRIPTION = 20
SYMBOL_PATH = 21

SYMBOL_TRADE_MODE_DISABLED = 0
SYMBOL_TRADE_MODE_LONGONLY = 3
SYMBOL_TRADE_MODE_SHORTONLY = 4
SYMBOL_TRADE_MODE_CLOSEONLY = 1
SYMBOL_TRADE_MODE_FULL = 2

SYMBOL_TRADE_EXECUTION_REQUEST = 0
SYMBOL_TRADE_EXECUTION_INSTANT = 1
SYMBOL_TRADE_EXECUTION_MARKET = 2
SYMBOL_TRADE_EXECUTION_EXCHANGE = 3

SUNDAY = 0
MONDAY = 1
TUESDAY = 2
WEDNESDAY = 3
THURSDAY = 4
FRIDAY = 5
SATURDAY = 6

# https://docs.mql4.com/constants/indicatorconstants/lines
MODE_MAIN = 0
MODE_SIGNAL = 1

MODE_PLUSDI = 1
MODE_MINUSDI = 2

MODE_UPPER = 1
MODE_LOWER = 2

MODE_GATORJAW = 1
MODE_GATORTEETH = 2
MODE_GATORLIPS = 3

MODE_TENKANSEN = 1
MODE_KIJUNSEN = 2
MODE_SENKOUSPANA = 3
MODE_SENKOUSPANB = 4
MODE_CHIKOUSPAN = 5

# https://docs.mql4.com/constants/tradingconstants/enum_series_info_integer
SERIES_BARS_COUNT = 0
SERIES_FIRSTDATE = 1
SERIES_LASTBAR_DATE = 5
SERIES_SERVER_FIRSTDATE = 2


# https://docs.mql4.com/constants/chartconstants/enum_timeframes
PERIOD_CURRENT = 0
PERIOD_M1 = 1
PERIOD_M5 = 5
PERIOD_M15 = 15
PERIOD_M30 = 30
PERIOD_H1 = 60
PERIOD_H4 = 240
PERIOD_D1 = 1440
PERIOD_W1 = 10080
PERIOD_MN1 = 43200

PERIOD_M2 = 2
PERIOD_M3 = 3
PERIOD_M4 = 4
PERIOD_M6 = 6
PERIOD_M10 = 10
PERIOD_M12 = 12
PERIOD_M20 = 20
PERIOD_H2 = 120
PERIOD_H3 = 180
PERIOD_H6 = 360
PERIOD_H8 = 480
PERIOD_H12 = 720

# https://docs.mql4.com/constants/indicatorconstants/enum_ma_method
MODE_SMA = 0
MODE_EMA = 1
MODE_SMMA = 2
MODE_LWMA = 3

# https://docs.mql4.com/constants/indicatorconstants/prices#enum_applied_price_enum
PRICE_CLOSE = 0
PRICE_OPEN = 1
PRICE_HIGH = 2
PRICE_LOW = 3
PRICE_MEDIAN = 4
PRICE_TYPICAL = 5
PRICE_WEIGHTED = 6

# # https://docs.mql4.com/constants/objectconstants/webcolors
clrNONE = ""
clrBlack = "0-0-0"
clrDarkGreen = "0-100-0"
clrDarkSlateGray = "47-79-79"
clrOlive = "128-128-0"
clrGreen = "0-128-0"
clrTeal = "0-128-128"
clrNavy = "0-0-128"
clrPurple = "128-0-128"
clrMaroon = "128-0-0"
clrIndigo = "75-0-130"
clrMidnightBlue = "25-25-112"
clrDarkBlue = "0-0-139"
clrDarkOliveGreen = "85-107-47"
clrSaddleBrown = "139-69-19"
clrForestGreen = "34-139-34"
clrOliveDrab = "107-142-35"
clrSeaGreen = "46-139-87"
clrDarkGoldenrod = "184-134-11"
clrDarkSlateBlue = "72-61-139"
clrSienna = "160-82-45"
clrMediumBlue = "0-0-205"
clrBrown = "165-42-42"
clrDarkTurquoise = "0-206-209"
clrDimGray = "105-105-105"
clrLightSeaGreen = "32-178-170"
clrDarkViolet = "148-0-211"
clrFireBrick = "178-34-34"
clrMediumVioletRed = "199-21-133"
clrMediumSeaGreen = "60-179-113"
clrChocolate = "210-105-30"
clrCrimson = "220-20-60"
clrSteelBlue = "70-130-180"
clrGoldenrod = "218-165-32"
clrMediumSpringGreen = "0-250-154"
clrLawnGreen = "124-252-0"
clrCadetBlue = "95-158-160"
clrDarkOrchid = "153-50-204"
clrYellowGreen = "154-205-50"
clrLimeGreen = "50-205-50"
clrOrangeRed = "255-69-0"
clrDarkOrange = "255-140-0"
clrOrange = "255-165-0"
clrGold = "255-215-0"
clrYellow = "255-255-0"
clrChartreuse = "127-255-0"
clrLime = "0-255-0"
clrSpringGreen = "0-255-127"
clrAqua = "0-255-255"
clrDeepSkyBlue = "0-191-255"
clrBlue = "0-0-255"
clrMagenta = "255-0-255"
clrRed = "255-0-0"
clrGray = "128-128-128"
clrSlateGray = "112-128-144"
clrPeru = "205-133-63"
clrBlueViolet = "138-43-226"
clrLightSlateGray = "119-136-153"
clrDeepPink = "255-20-147"
clrMediumTurquoise = "72-209-204"
clrDodgerBlue = "30-144-255"
clrTurquoise = "64-224-208"
clrRoyalBlue = "65-105-225"
clrSlateBlue = "106-90-205"
clrDarkKhaki = "189-183-107"
clrIndianRed = "205-92-92"
clrMediumOrchid = "186-85-211"
clrGreenYellow = "173-255-47"
clrMediumAquamarine = "102-205-170"
clrDarkSeaGreen = "143-188-143"
clrTomato = "255-99-71"
clrRosyBrown = "188-143-143"
clrOrchid = "218-112-214"
clrMediumPurple = "147-112-219"
clrPaleVioletRed = "219-112-147"
clrCoral = "255-127-80"
clrCornflowerBlue = "100-149-237"
clrDarkGray = "169-169-169"
clrSandyBrown = "244-164-96"
clrMediumSlateBlue = "123-104-238"
clrTan = "210-180-140"
clrDarkSalmon = "233-150-122"
clrBurlyWood = "222-184-135"
clrHotPink = "255-105-180"
clrSalmon = "250-128-114"
clrViolet = "238-130-238"
clrLightCoral = "240-128-128"
clrSkyBlue = "135-206-235"
clrLightSalmon = "255-160-122"
clrPlum = "221-160-221"
clrKhaki = "240-230-140"
clrLightGreen = "144-238-144"
clrAquamarine = "127-255-212"
clrSilver = "192-192-192"
clrLightSkyBlue = "135-206-250"
clrLightSteelBlue = "176-196-222"
clrLightBlue = "173-216-230"
clrPaleGreen = "152-251-152"
clrThistle = "216-191-216"
clrPowderBlue = "176-224-230"
clrPaleGoldenrod = "238-232-170"
clrPaleTurquoise = "175-238-238"
clrLightGray = "211-211-211"
clrWheat = "245-222-179"
clrNavajoWhite = "255-222-173"
clrMoccasin = "255-228-181"
clrLightPink = "255-182-193"
clrGainsboro = "220-220-220"
clrPeachPuff = "255-218-185"
clrPink = "255-192-203"
clrBisque = "255-228-196"
clrLightGoldenrod = "250-250-210"
clrBlanchedAlmond = "255-235-205"
clrLemonChiffon = "255-250-205"
clrBeige = "245-245-220"
clrAntiqueWhite = "250-235-215"
clrPapayaWhip = "255-239-213"
clrCornsilk = "255-248-220"
clrLightYellow = "255-255-224"
clrLightCyan = "224-255-255"
clrLinen = "250-240-230"
clrLavender = "230-230-250"
clrMistyRose = "255-228-225"
clrOldLace = "253-245-230"
clrWhiteSmoke = "245-245-245"
clrSeashell = "255-245-238"
clrIvory = "255-255-240"
clrHoneydew = "240-255-240"
clrAliceBlue = "240-248-255"
clrLavenderBlush = "255-240-245"
clrMintCream = "245-255-250"
clrSnow = "255-250-250"
clrWhite = "255-255-255"



from enum import Enum
from typing import Union, Set, List, Tuple, Dict, Any
import threading
import signal
import traceback
import socket
import select

class TCPStatus(Enum):
    SUCCESS = 1
    ERROR = 2
    READ = 3
    TIMEOUT = 4

class TCPCast(object):
    def __init__(self) -> None:
        self.__socket: socket = None
        self.__if_address: str = ""
        self.__if_port: int = 0
        self.__is_client: bool = False
        self.__num_clients: int = 0
        self.__buffer_size: int = 1024 * 1024 * 5 # 5 MB
        self.__client_mutex: threading.Lock = threading.Lock()
        self.__server_mutex: threading.Lock = threading.Lock()
        self.__timeout_in_seconds: int = 1
        self.__is_connected: bool = False

        signal.signal(signal.SIGINT, self.__stop)

    def InitComponent(self
                      , if_address: str
                      , if_port: int
                      , is_client: bool
                      , num_clients: int = 5) -> TCPStatus:
        self.__if_address = if_address
        self.__if_port = if_port
        self.__is_client = is_client
        self.__num_clients = num_clients

        if self.__is_client:
            self.__if_port = 0
            self.__num_clients = 0

        return self.__start(if_address=self.__if_address
                            , if_port=self.__if_port
                            , is_client=self.__is_client
                            , num_clients=self.__num_clients)

    # Client
    def Connect(self, to_address: str, to_port: int) -> TCPStatus:
        try:
            self.__socket.connect((to_address, to_port))
            with self.__client_mutex:
                self.__is_connected = True
        except Exception as e:
            with self.__client_mutex:
                self.__is_connected = False
            return TCPStatus.ERROR

        return TCPStatus.SUCCESS

    def ClientSend(self, send_msg: str) -> TCPStatus:
        with self.__client_mutex:
            try:
                self.__socket.send(send_msg.encode())
            except Exception as e:
                self.__is_connected = False
                return TCPStatus.ERROR

        return TCPStatus.SUCCESS

    def ClientRecv(self) -> Union[TCPStatus, str]:
        try:
            ready = select.select([self.__socket], [], [], self.__timeout_in_seconds)
            if ready[0]:
                msg = self.__socket.recv(self.__buffer_size)
                return (TCPStatus.SUCCESS, msg.decode())
            return (TCPStatus.TIMEOUT, None)
        except Exception as e:
            with self.__client_mutex:
                self.__is_connected = False
            return (TCPStatus.ERROR, None)

    def IsClientConnected(self) -> bool:
        with self.__client_mutex:
            return self.__is_connected

        return False

    # Server
    def Accept(self) -> Union[socket.socket, str]:
        try:
            client_socket, client_address = self.__socket.accept()
            return (client_socket, client_address)
        except Exception as e:
            return (None, None)

    def ServerSend(self, client_socket: socket.socket, send_msg: str) -> TCPStatus:
        try:
            with self.__server_mutex:
                client_socket.send(send_msg.encode())
            return TCPStatus.SUCCESS
        except Exception as e:
            return TCPStatus.ERROR

    def ServerRecv(self, client_socket: socket.socket) -> Union[TCPStatus, str]:
        try:
            ready = select.select([client_socket], [], [], self.__timeout_in_seconds)
            if ready[0]:
                msg = client_socket.recv(self.__buffer_size)
                return (TCPStatus.SUCCESS, msg.decode())
            return (TCPStatus.TIMEOUT, None)
        except Exception as e:
            return (TCPStatus.ERROR, None)

    def __start(self
                , if_address: str
                , if_port: int
                , is_client: bool
                , num_clients) -> TCPStatus:
        self.__socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if self.__socket == None:
            return TCPStatus.ERROR

        if is_client == False:
            self.__socket.bind((if_address, if_port))
            self.__socket.listen(num_clients)

        # self.__socket.setblocking(0)
        return TCPStatus.SUCCESS

    def __stop(self, signum, frame):
        if self.__socket is not None:
            self.__socket.shutdown(how=0)



from typing import Dict, List, Set, Tuple, Union, Any
from datetime import datetime

import pysailfish.internal.MT_EA.mt4_const as mc
from pysailfish.internal.DataCtrl.MT4DataCtrl.MT4DataCtrl import MT4DataCtrl

class timeseries_and_indicators_access(object):
    def __init__(self):
        self.__data_ctrl: MT4DataCtrl = None
        self.__fun_num_map: Dict[str, int] = dict()

    def init_component(self, data_ctrl: MT4DataCtrl, fun_num_map: Dict[str, int]):
        self.__data_ctrl = data_ctrl
        self.__fun_num_map = fun_num_map

    def series_info_integer(self, symbol_name: str, timeframe: int, prop_id: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["SeriesInfoInteger"]))
        self.__data_ctrl.add_data_to_msg(data=symbol_name)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(prop_id))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def copy_rates_001(self, symbol_name: str, timeframe: int, start_pos: int, count: int) -> List[Union[datetime, float, float, float, float, int, int, int]]:
        return self.__copy_001(symbol_name=symbol_name
                               , timeframe=timeframe
                               , start_pos=start_pos
                               , count=count
                               , fun_name="CopyRates_001"
                               , parse_fun=self.__copy_rates_common)

    def copy_rates_002(self, symbol_name: str, timeframe: int, start_time: datetime, count: int) -> List[Union[datetime, float, float, float, float, int, int, int]]:
        return self.__copy_002(symbol_name=symbol_name
                               , timeframe=timeframe
                               , start_time=start_time
                               , count=count
                               , fun_name="CopyRates_002"
                               , parse_fun=self.__copy_rates_common)

    def copy_rates_003(self, symbol_name: str, timeframe: int, start_time: datetime, end_time: datetime) -> List[Union[datetime, float, float, float, float, int, int, int]]:
        return self.__copy_003(symbol_name=symbol_name
                               , timeframe=timeframe
                               , start_time=start_time
                               , end_time=end_time
                               , fun_name="CopyRates_003"
                               , parse_fun=self.__copy_rates_common)

    def copy_time_001(self, symbol_name: str, timeframe: int, start_pos: int, count: int) -> List[datetime]:
        return self.__copy_001(symbol_name=symbol_name
                               , timeframe=timeframe
                               , start_pos=start_pos
                               , count=count
                               , fun_name="CopyTime_001"
                               , parse_fun=self.__copy_time_common)

    def copy_time_002(self, symbol_name: str, timeframe: int, start_time: datetime, count: int) -> List[datetime]:
        return self.__copy_002(symbol_name=symbol_name
                               , timeframe=timeframe
                               , start_time=start_time
                               , count=count
                               , fun_name="CopyTime_002"
                               , parse_fun=self.__copy_time_common)

    def copy_time_003(self, symbol_name: str, timeframe: int, start_time: datetime, end_time: datetime) -> List[datetime]:
        return self.__copy_003(symbol_name=symbol_name
                               , timeframe=timeframe
                               , start_time=start_time
                               , end_time=end_time
                               , fun_name="CopyTime_003"
                               , parse_fun=self.__copy_time_common)

    def copy_open_001(self, symbol_name: str, timeframe: int, start_pos: int, count: int) -> List[float]:
        return self.__copy_001(symbol_name=symbol_name
                               , timeframe=timeframe
                               , start_pos=start_pos
                               , count=count
                               , fun_name="CopyOpen_001"
                               , parse_fun=self.__copy_float_common)

    def copy_open_002(self, symbol_name: str, timeframe: int, start_time: datetime, count: int) -> List[float]:
        return self.__copy_002(symbol_name=symbol_name
                               , timeframe=timeframe
                               , start_time=start_time
                               , count=count
                               , fun_name="CopyOpen_002"
                               , parse_fun=self.__copy_float_common)

    def copy_open_003(self, symbol_name: str, timeframe: int, start_time: datetime, end_time: datetime) -> List[float]:
        return self.__copy_003(symbol_name=symbol_name
                               , timeframe=timeframe
                               , start_time=start_time
                               , end_time=end_time
                               , fun_name="CopyOpen_003"
                               , parse_fun=self.__copy_float_common)

    def copy_high_001(self, symbol_name: str, timeframe: int, start_pos: int, count: int) -> List[float]:
        return self.__copy_001(symbol_name=symbol_name
                               , timeframe=timeframe
                               , start_pos=start_pos
                               , count=count
                               , fun_name="CopyHigh_001"
                               , parse_fun=self.__copy_float_common)

    def copy_high_002(self, symbol_name: str, timeframe: int, start_time: datetime, count: int) -> List[float]:
        return self.__copy_002(symbol_name=symbol_name
                               , timeframe=timeframe
                               , start_time=start_time
                               , count=count
                               , fun_name="CopyHigh_002"
                               , parse_fun=self.__copy_float_common)

    def copy_high_003(self, symbol_name: str, timeframe: int, start_time: datetime, end_time: datetime) -> List[float]:
        return self.__copy_003(symbol_name=symbol_name
                               , timeframe=timeframe
                               , start_time=start_time
                               , end_time=end_time
                               , fun_name="CopyHigh_003"
                               , parse_fun=self.__copy_float_common)

    def copy_low_001(self, symbol_name: str, timeframe: int, start_pos: int, count: int) -> List[float]:
        return self.__copy_001(symbol_name=symbol_name
                               , timeframe=timeframe
                               , start_pos=start_pos
                               , count=count
                               , fun_name="CopyLow_001"
                               , parse_fun=self.__copy_float_common)

    def copy_low_002(self, symbol_name: str, timeframe: int, start_time: datetime, count: int) -> List[float]:
        return self.__copy_002(symbol_name=symbol_name
                               , timeframe=timeframe
                               , start_time=start_time
                               , count=count
                               , fun_name="CopyLow_002"
                               , parse_fun=self.__copy_float_common)

    def copy_low_003(self, symbol_name: str, timeframe: int, start_time: datetime, end_time: datetime) -> List[float]:
        return self.__copy_003(symbol_name=symbol_name
                               , timeframe=timeframe
                               , start_time=start_time
                               , end_time=end_time
                               , fun_name="CopyLow_003"
                               , parse_fun=self.__copy_float_common)

    def copy_close_001(self, symbol_name: str, timeframe: int, start_pos: int, count: int) -> List[float]:
        return self.__copy_001(symbol_name=symbol_name
                               , timeframe=timeframe
                               , start_pos=start_pos
                               , count=count
                               , fun_name="CopyClose_001"
                               , parse_fun=self.__copy_float_common)

    def copy_close_002(self, symbol_name: str, timeframe: int, start_time: datetime, count: int) -> List[float]:
        return self.__copy_002(symbol_name=symbol_name
                               , timeframe=timeframe
                               , start_time=start_time
                               , count=count
                               , fun_name="CopyClose_002"
                               , parse_fun=self.__copy_float_common)

    def copy_close_003(self, symbol_name: str, timeframe: int, start_time: datetime, end_time: datetime) -> List[float]:
        return self.__copy_003(symbol_name=symbol_name
                               , timeframe=timeframe
                               , start_time=start_time
                               , end_time=end_time
                               , fun_name="CopyClose_003"
                               , parse_fun=self.__copy_float_common)

    def copy_tick_volume_001(self, symbol_name: str, timeframe: int, start_pos: int, count: int) -> List[int]:
        return self.__copy_001(symbol_name=symbol_name
                               , timeframe=timeframe
                               , start_pos=start_pos
                               , count=count
                               , fun_name="CopyTickVolume_001"
                               , parse_fun=self.__copy_int_common)

    def copy_tick_volume_002(self, symbol_name: str, timeframe: int, start_time: datetime, count: int) -> List[int]:
        return self.__copy_002(symbol_name=symbol_name
                               , timeframe=timeframe
                               , start_time=start_time
                               , count=count
                               , fun_name="CopyTickVolume_002"
                               , parse_fun=self.__copy_int_common)

    def copy_tick_volume_003(self, symbol_name: str, timeframe: int, start_time: datetime, end_time: datetime) -> List[int]:
        return self.__copy_003(symbol_name=symbol_name
                               , timeframe=timeframe
                               , start_time=start_time
                               , end_time=end_time
                               , fun_name="CopyTickVolume_003"
                               , parse_fun=self.__copy_int_common)

    def __copy_001(self, symbol_name: str, timeframe: int, start_pos: int, count: int, fun_name: str, parse_fun: Any) -> List[Any]:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map[fun_name]))
        self.__data_ctrl.add_data_to_msg(data=symbol_name)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(start_pos))
        self.__data_ctrl.add_data_to_msg(data=str(count))
        self.__data_ctrl.send_message()

        return parse_fun(str_ret=self.__data_ctrl.receive_string_from_mt4())

    def __copy_002(self, symbol_name: str, timeframe: int, start_time: datetime, count: int, fun_name: str, parse_fun: Any) -> List[Any]:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map[fun_name]))
        self.__data_ctrl.add_data_to_msg(data=symbol_name)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=start_time.strftime("%Y.%m.%d %H:%M:%S"))
        self.__data_ctrl.add_data_to_msg(data=str(count))
        self.__data_ctrl.send_message()

        return parse_fun(str_ret=self.__data_ctrl.receive_string_from_mt4())

    def __copy_003(self, symbol_name: str, timeframe: int, start_time: datetime, end_time: datetime, fun_name: str, parse_fun: Any) -> List[Any]:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map[fun_name]))
        self.__data_ctrl.add_data_to_msg(data=symbol_name)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=start_time.strftime("%Y.%m.%d %H:%M:%S"))
        self.__data_ctrl.add_data_to_msg(data=end_time.strftime("%Y.%m.%d %H:%M:%S"))
        self.__data_ctrl.send_message()

        return parse_fun(str_ret=self.__data_ctrl.receive_string_from_mt4())

    def __copy_time_common(self, str_ret: str) -> List[datetime]:
        if str_ret == "-1":
            return list()
        parts = str_ret.split("~")
        arr_size = int(parts[0])
        if arr_size+1 != len(parts):
            return list()

        datetime_array: List[datetime] = list()
        for ele in parts[1:]:
            datetime_array.append(datetime.strptime(ele, "%Y.%m.%d %H:%M:%S"))

        return datetime_array

    def __copy_float_common(self, str_ret: str) -> List[float]:
        if str_ret == "-1":
            return list()
        parts = str_ret.split("~")
        arr_size = int(parts[0])
        if arr_size+1 != len(parts):
            return list()

        float_array: List[float] = list()
        for ele in parts[1:]:
            float_array.append(float(ele))

        return float_array

    def __copy_int_common(self, str_ret: str) -> List[int]:
        if str_ret == "-1":
            return list()
        parts = str_ret.split("~")
        arr_size = int(parts[0])
        if arr_size+1 != len(parts):
            return list()

        int_array: List[float] = list()
        for ele in parts[1:]:
            int_array.append(int(ele))

        return int_array

    def __copy_rates_common(self, str_ret: str) -> List[Union[datetime, float, float, float, float, int, int, int]]:
        if str_ret == "-1":
            return list()
        parts = str_ret.split("~")
        arr_size = int(parts[0])
        if (arr_size*8)+1 != len(parts):
            return list()

        rates_array: List[Union[datetime, float, float, float, float, int, int, int]] = list()
        for idx in range(1, len(parts), 8):
            rates_array.append((datetime.strptime(parts[idx], "%Y.%m.%d %H:%M:%S")
                                , float(parts[idx+1])
                                , float(parts[idx+2])
                                , float(parts[idx+3])
                                , float(parts[idx+4])
                                , int(parts[idx+5])
                                , int(parts[idx+6])
                                , int(parts[idx+7])))

        return rates_array

    def bars_001(self, symbol_name: str, timeframe: int) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["Bars_001"]))
        self.__data_ctrl.add_data_to_msg(data=symbol_name)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_int_from_mt4()

    def bars_002(self, symbol_name: str, timeframe: int, start_time: datetime, stop_time: datetime) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["Bars_002"]))
        self.__data_ctrl.add_data_to_msg(data=symbol_name)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=start_time.strftime("%Y.%m.%d %H:%M:%S"))
        self.__data_ctrl.add_data_to_msg(data=stop_time.strftime("%Y.%m.%d %H:%M:%S"))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_int_from_mt4()

    def ibars(self, symbol: str, timeframe: int) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iBars"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_int_from_mt4()

    def ibar_shift(self, symbol: str, timeframe: int, time: datetime, exact: bool) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iBarShift"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=time.strftime("%Y.%m.%d %H:%M:%S"))
        self.__data_ctrl.add_data_to_msg(data=str(exact))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_int_from_mt4()

    def iclose(self, symbol: str, timeframe: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iClose"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def ihigh(self, symbol: str, timeframe: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iHigh"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def ihighest(self, symbol: str, timeframe: int, type: int, count: int, start: int) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iHighest"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(type))
        self.__data_ctrl.add_data_to_msg(data=str(count))
        self.__data_ctrl.add_data_to_msg(data=str(start))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_int_from_mt4()

    def ilow(self, symbol: str, timeframe: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iLow"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def ilowest(self, symbol: str, timeframe: int, type: int, count: int, start: int) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iLowest"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(type))
        self.__data_ctrl.add_data_to_msg(data=str(count))
        self.__data_ctrl.add_data_to_msg(data=str(start))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_int_from_mt4()

    def iopen(self, symbol: str, timeframe: int, shift: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iOpen"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_float_from_mt4()

    def itime(self, symbol: str, timeframe: int, shift: int) -> datetime:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iTime"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_datetime_from_mt4()

    def ivolume(self, symbol: str, timeframe: int, shift: int) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["iVolume"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(timeframe))
        self.__data_ctrl.add_data_to_msg(data=str(shift))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_int_from_mt4()



from typing import Dict, List, Set, Tuple, Union, Any
from datetime import datetime

import pysailfish.internal.MT_EA.mt4_const as mc
from pysailfish.internal.DataCtrl.MT4DataCtrl.MT4DataCtrl import MT4DataCtrl

# https://docs.mql4.com/objects/objectname
class chart_operations(object):
    def __init__(self):
        self.__data_ctrl: MT4DataCtrl = None
        self.__fun_num_map: Dict[str, int] = dict()

    def init_component(self, data_ctrl: MT4DataCtrl, fun_num_map: Dict[str, int]):
        self.__data_ctrl = data_ctrl
        self.__fun_num_map = fun_num_map

    def chart_apply_template(self, **kwargs) -> bool:
        return None

    def chart_save_template(self, **kwargs) -> bool:
        return None

    def chart_window_find_001(self, **kwargs) -> int:
        return None

    def chart_window_find_002(self, **kwargs) -> int:
        return None

    def chart_time_price_to_xy(self, **kwargs) -> bool:
        return None

    def chart_xy_to_time_price(self, **kwargs) -> bool:
        return None

    def chart_open(self, **kwargs) -> int:
        return None

    def chart_first(self) -> int:
        return None

    def chart_next(self, **kwargs) -> int:
        return None

    def chart_close(self, **kwargs) -> bool:
        return None

    def chart_symbol(self, **kwargs) -> str:
        return None

    def chart_period(self, **kwargs) -> int:
        return None

    def chart_redraw(self, **kwargs) -> None:
        return None

    def chart_set_double(self, **kwargs) -> bool:
        return None

    def chart_set_integer_001(self, **kwargs) -> bool:
        return None

    def chart_set_integer_002(self, **kwargs) -> bool:
        return None

    def chart_set_string(self, **kwargs) -> bool:
        return None

    def chart_get_double_001(self, **kwargs) -> float:
        return None

    def chart_get_double_002(self, **kwargs) -> bool:
        return None

    def chart_get_integer_001(self, **kwargs) -> int:
        return None

    def chart_get_integer_002(self, **kwargs) -> bool:
        return None

    def chart_get_string_001(self, **kwargs) -> str:
        return None

    def chart_get_string_002(self, **kwargs) -> bool:
        return None

    def chart_navigate(self, **kwargs) -> bool:
        return None

    def chart_id(self) -> int:
        return None

    def chart_indicator_delete(self, **kwargs) -> bool:
        return None

    def chart_indicator_name(self, **kwargs) -> str:
        return None

    def chart_indicators_total(self, **kwargs) -> int:
        return None

    def chart_window_on_dropped(self) -> int:
        return None

    def chart_price_on_dropped(self) -> float:
        return None

    def chart_time_on_dropped(self) -> datetime:
        return None

    def chart_x_on_dropped(self) -> int:
        return None

    def chart_y_on_dropped(self) -> int:
        return None

    def chart_set_symbol_period(self, chart_id: int, symbol: str, period: int) -> bool:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["ChartSetSymbolPeriod"]))
        self.__data_ctrl.add_data_to_msg(data=str(chart_id))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=str(period))
        self.__data_ctrl.send_message()

        return self.__data_ctrl.receive_bool_from_mt4()

    def chart_screen_shot(self, **kwargs) -> bool:
        return None

    def period(self) -> int:
        return None

    def symbol(self) -> str:
        return None

    def window_bars_per_chart(self) -> int:
        return None

    def window_expert_name(self) -> str:
        return None

    def window_find(self, **kwargs) -> int:
        return None



from typing import Dict, List, Set, Tuple, Union, Any

import pysailfish.internal.MT_EA.mt4_const as mc
from pysailfish.internal.DataCtrl.MT4DataCtrl.MT4DataCtrl import MT4DataCtrl

class account_information(object):
    def __init__(self):
        self.__data_ctrl: MT4DataCtrl = None
        self.__fun_num_map: Dict[str, int] = dict()

    def init_component(self, data_ctrl: MT4DataCtrl, fun_num_map: Dict[str, int]):
        self.__data_ctrl = data_ctrl
        self.__fun_num_map = fun_num_map

    def account_info_double(self, property_id: int) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["AccountInfoDouble"]))
        self.__data_ctrl.add_data_to_msg(data=str(property_id))

        return self.__data_ctrl.receive_float_from_mt4()

    def account_info_integer(self, property_id: int) -> int:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["AccountInfoInteger"]))
        self.__data_ctrl.add_data_to_msg(data=str(property_id))

        return self.__data_ctrl.receive_int_from_mt4()

    def account_info_string(self, property_id: int) -> str:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["AccountInfoString"]))
        self.__data_ctrl.add_data_to_msg(data=str(property_id))

        return self.__data_ctrl.receive_string_from_mt4()

    def account_free_margin_check(self, symbol: str, cmd: int, volume: float) -> float:
        self.__data_ctrl.add_data_to_msg(data=str(self.__fun_num_map["AccountFreeMarginCheck"]))
        self.__data_ctrl.add_data_to_msg(data=symbol)
        self.__data_ctrl.add_data_to_msg(data=int(cmd))
        self.__data_ctrl.add_data_to_msg(data=int(volume))

        return self.__data_ctrl.receive_float_from_mt4()
